/*
*************************************************************************************************
**	O.S		: Linux
**	FILE NAME	: osm.h
**	BY		: Nick Cheng   
**	Description	: SCSI Device Driver for Areca Host Bus Adapter 
*******************************************************************************
** Copyright (C) 2002 - 2005, Areca Technology Corporation All rights reserved.
**
**	Web site: www.areca.com.tw
**	E-mail: support@areca.com.tw
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
** This program is distributed in the hope that it will be useful, but 
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
*******************************************************************************
** Redistribution and use in source and binary forms,with or without
** modification,are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**    notice,this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice,this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
** 3. The name of the author may not be used to endorse or promote products
**    derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
** IMPLIED WARRANTIES,INCLUDING,BUT NOT LIMITED TO,THE IMPLIED WARRANTIES
** OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
** IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,INDIRECT,
** INCIDENTAL,SPECIAL,EXEMPLARY,OR CONSEQUENTIAL DAMAGES(INCLUDING,BUT
** NOT LIMITED TO,PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA,OR PROFITS; OR BUSINESS INTERRUPTION)HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY,WHETHER IN CONTRACT,STRICT LIABILITY,OR TORT
**(INCLUDING NEGLIGENCE OR OTHERWISE)ARISING IN ANY WAY OUT OF THE USE OF
** THIS SOFTWARE,EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*********************************************************************************
** History
**
** REV#			DATE		NAME				DESCRIPTION
** 1.1.0.90505		05/05/2009	Nick Cheng
*************************************************************************************************
*/
//#define _DEBUG_ALL
#define _DEBUG_TMT
#if defined __KERNEL__
	#include <linux/version.h>
	#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
		#define MODVERSIONS
	#endif
	/* modversions.h should be before should be before module.h */
	#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
		#if defined(MODVERSIONS)
			#include <config/modversions.h>
		#endif
	#endif
	#include <linux/kernel.h>
	#include <scsi/scsi_host.h>
	#include <linux/pci.h>
	#include <linux/device.h>
	#include <linux/dmapool.h>
	#include <linux/dma-mapping.h>
	#include <linux/sched.h>
	#include <linux/jiffies.h>
	#include <linux/module.h>
	#include <asm/dma.h>
	#include <asm/io.h>
	#include <asm/system.h>
	#include <asm/uaccess.h>
	#include <linux/delay.h>
	#include <linux/signal.h>
	#include <linux/errno.h>
	#include <linux/ioport.h>
	#include <linux/proc_fs.h>
	#include <linux/string.h>
	#include <linux/ctype.h>
	#include <linux/interrupt.h>
	//#include <linux/smp_lock.h>
	#include <linux/stddef.h>
	#include <linux/pci_ids.h>
	#include <linux/moduleparam.h>
	#include <linux/blkdev.h>
	#include <linux/reboot.h>
	#include <linux/sched.h>
	#include <linux/init.h>
	#include <linux/highmem.h>
	#include <linux/spinlock.h>
	#include <linux/nmi.h>
	#include <linux/kthread.h>
	#include <linux/rtc.h>
	#include <asm/rtc.h>
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 14)
		#include <linux/pci_regs.h>
		#include<linux/sched.h>
	#endif
	#include <scsi/scsi.h>
	#include <scsi/scsi_device.h>
	#include <scsi/scsi_cmnd.h>
	#include <scsi/scsi_tcq.h>
	#include <scsi/scsi_device.h>
	#include <scsi/scsicam.h>
	#include <scsi/scsi_eh.h>
	#include <linux/string.h>
	#include <linux/workqueue.h>
#endif
#ifdef _DEBUG_ALL
	#define ARCSAS_DBG_FUNCTION			1
	#define ARCSAS_DBG_CONTEXT			1
	#define ARCSAS_DBG_LINKEDLIST			1
	#define ARCSAS_DBG_OS_CCB_CONTEXT		1
	#define ARCSAS_DBG_OS_CCB			1
	#define ARCSAS_DBG_INTERRUPT			1
	#define ARCSAS_DBG_DISCOVERY			1
	#define ARCSAS_DBG_DISCOVERY_CONTEXT		1
	#define ARCSAS_DBG_SATA_SM			1
	#define ARCSAS_DBG_SATA_SM_CONTEXT		1
	#define ARCSAS_DBG_SAS_SM			1
	#define ARCSAS_DBG_SAS_SM_CONTEXT		1
	#define ARCSAS_DBG_PORT_SM			1
	#define ARCSAS_DBG_PORT_SM_CONTEXT		1
	#define ARCSAS_DBG_SES				1
	#define ARCSAS_DBG_SES_CONTEXT			1
	#define ARCSAS_DBG_HAS_SENSE			1
	#define ARCSAS_DBG_IOCTL			1
	#define ARCSAS_DBG_IOCTL_CONTEXT		1
#else
	#define ARCSAS_DBG_FUNCTION			0
	#define ARCSAS_DBG_CONTEXT			0
	#define ARCSAS_DBG_LINKEDLIST			0
	#define ARCSAS_DBG_OS_CCB_CONTEXT		0
	#define ARCSAS_DBG_OS_CCB			0			
	#define ARCSAS_DBG_INTERRUPT			0
	#define ARCSAS_DBG_DISCOVERY			0
	#define ARCSAS_DBG_DISCOVERY_CONTEXT		0
	#define ARCSAS_DBG_SATA_SM			0
	#define ARCSAS_DBG_SATA_SM_CONTEXT		0
	#define ARCSAS_DBG_SAS_SM			0
	#define ARCSAS_DBG_SAS_SM_CONTEXT		0
	#define ARCSAS_DBG_PORT_SM			0
	#define ARCSAS_DBG_PORT_SM_CONTEXT		0
	#define ARCSAS_DBG_SES				0
	#define ARCSAS_DBG_SES_CONTEXT			0
	#define ARCSAS_DBG_HAS_SENSE			0
	#define ARCSAS_DBG_IOCTL			0
	#define ARCSAS_DBG_IOCTL_CONTEXT		0
#endif

#ifdef  _DEBUG_TMT
	#define ARCSAS_DBG_TIMEOUT			1
	#define ARCSAS_DBG_TASKFILE_ERR		1
	#define ARCSAS_DBG_OS_EH			1
	#define ARCSAS_DBG_PCI_REMOVE			1
	#define ARCSAS_DBG_PLUG_UNPLUG		1
	#define ARCSAS_DBG_PORT_ABORT_CCB		1
	#define ARCSAS_DBG_SLOT_RESET			1
#endif

#define ARCSAS_SUPPORT_SES				0
#define ARCSAS_SUPPORT_SGPIO				0
#define ARCSAS_SUPPORT_IOCTL				0
#if ARCSAS_SUPPORT_SGPIO
	#define ARCSAS_SUPPORT_SGPIO_TEST			1
	#define ARCSAS_SUPPORT_SGPIO_SDATA_IN_OUT		1
	#define ARCSAS_SUPPORT_GUI_SGPIO_REG_READ_WRITE	1
#else
	#define ARCSAS_SUPPORT_SGPIO_TEST			0
	#define ARCSAS_SUPPORT_SGPIO_SDATA_IN_OUT		0
	#define ARCSAS_SUPPORT_GUI_SGPIO_REG_READ_WRITE	0
#endif
#if ARCSAS_SUPPORT_IOCTL	
	#define ARCSAS_SUPPORT_SMART			1
#else
	#define ARCSAS_SUPPORT_SMART			0
#endif
extern unsigned char modePageBuf[];
extern unsigned char info_except_ctrl_mode_page[];
extern struct AdapterControlBlock *acb_addr;
extern struct timezone sys_tz;
typedef dma_addr_t osm_dma_addr_t;

#ifndef offsetof
	#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#endif

#ifndef container_of
#define container_of(ptr, type, member) ({          \
	const typeof(((type *)0)->member)*__mptr = (ptr);    \
		     (type *)((char *)__mptr - offsetof(type, member)); })
#endif
/*
**************************************************************************************
**************************************************************************************
*/
#define ARCSAS_SWAP_32(x)	(((u32)((unsigned char)(x)))<<24 | ((u32)((unsigned char)((x)>>8)))<<16 | ((u32)((unsigned char)((x)>>16)))<<8 | ((u32)((unsigned char)((x)>>24))) )
#define ARCSAS_SWAP_64(x)	((((unsigned long long)) (ARCSAS_SWAP_32((x).parts.low))) << 32 | ARCSAS_SWAP_32((x).parts.high))
#define ARCSAS_SWAP_16(x)	(((u16) ((unsigned char) (x))) << 8 | (u16) ((unsigned char) ((x) >> 8)))
#ifdef _BIG_ENDIAN__
	#define ARCSAS_CPU_TO_LE16_PTR(pu16)	*((unsigned short *)(pu16)) = ARCSAS_CPU_TO_LE16(*(unsigned short *)(pu16))	/*cpu_to_le16p(pu16)*/
	#define ARCSAS_CPU_TO_LE32_PTR(pu32)	*((unsigned int *)(pu32)) = ARCSAS_CPU_TO_LE32(*(unsigned int *)(pu32))		/*cpu_to_le32p(pu32)*/
	#define ARCSAS_LE16_TO_CPU_PTR(pu16)	*((unsigned short *)(pu16)) = ARCSAS_LE16_TO_CPU(*(unsigned short *)(pu16))	/*le16_to_cpup(pu16)*/
	#define ARCSAS_LE32_TO_CPU_PTR(pu32)	*((unsigned int *)(pu32)) = ARCSAS_LE32_TO_CPU(*(unsigned int *)(pu32))		/*le32_to_cpup(pu32)*/
#else
	#define ARCSAS_CPU_TO_LE16_PTR(pu16)        
	#define ARCSAS_CPU_TO_LE32_PTR(pu32)        
	#define ARCSAS_LE16_TO_CPU_PTR(pu32)
	#define ARCSAS_LE32_TO_CPU_PTR(pu32)
#endif
	#define ARCSAS_CPU_TO_LE16		cpu_to_le16
	#define ARCSAS_CPU_TO_LE32		cpu_to_le32
	#define ARCSAS_CPU_TO_LE64(x)		cpu_to_le64((x).value)
	#define ARCSAS_CPU_TO_BE16		cpu_to_be16
	#define ARCSAS_CPU_TO_BE32		cpu_to_be32
	#define ARCSAS_CPU_TO_BE64(x)		cpu_to_be64((x).value)

	#define ARCSAS_LE16_TO_CPU		le16_to_cpu
	#define ARCSAS_LE32_TO_CPU		le32_to_cpu
	#define ARCSAS_LE64_TO_CPU(x)		le64_to_cpu((x).value)
	#define ARCSAS_BE16_TO_CPU		be16_to_cpu
	#define ARCSAS_BE32_TO_CPU		be32_to_cpu
	#define ARCSAS_BE64_TO_CPU(x)		be64_to_cpu((x).value)
/*
**************************************************************************************
**************************************************************************************
*/
#ifndef SCSI_DEFAULT_MAX_SECTORS
	#define SCSI_DEFAULT_MAX_SECTORS	1024
#endif
/*
**************************************************************************************
**************************************************************************************
*/
#ifndef DMA_BIT_MASK
	#define DMA_BIT_MASK(n)	(((n) == 64) ? ~0ULL : ((1ULL<<(n))-1))
#endif
/*
**************************************************************************************
**        split 64bits dma addressing
**************************************************************************************
*/
#define dma_addr_hi32(addr)		(uint32_t) ((addr>>16)>>16)
#define dma_addr_lo32(addr) 		(uint32_t) (addr & 0xffffffff)
/*
**************************************************************************************
**************************************************************************************
*/
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 24)
	#define arcsas_use_sg(cmd)	scsi_sg_count(cmd)
	#define arcsas_rq_bf(cmd)		scsi_sglist(cmd)
	#define arcsas_rq_bflen(cmd)	scsi_bufflen(cmd)
	#define arcsas_map_sg(cmd)	scsi_dma_map(cmd)
	#define arcsas_unmap_sg(cmd)	scsi_dma_unmap(cmd)
	#define arcsas_chg_bflen(cmd)	cmd->sdb.length
					
#else
	#define arcsas_use_sg(cmd)	cmd->use_sg
	#define arcsas_rq_bf(cmd)		cmd->request_buffer
	#define arcsas_rq_bflen(cmd)	cmd->request_bufflen
	#define arcsas_map_sg(cmd)	pci_map_sg(pACB->pcidev, arcsas_rq_bf(cmd), arcsas_use_sg(cmd), cmd->sc_data_direction)
	#define arcsas_unmap_sg(cmd)	pci_unmap_sg(pACB->pcidev, arcsas_rq_bf(cmd), arcsas_use_sg(cmd), cmd->sc_data_direction)
	#define arcsas_chg_bflen(cmd)	cmd->request_bufflen
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 24)
	#define arcsas_map_sg_page(sg)	kmap_atomic(sg->page, KM_IRQ0)
#else
	#define arcsas_map_sg_page(sg)	kmap_atomic(sg_page(sg), KM_IRQ0)
#endif

#if !defined(scsi_sglist) && LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 24)
	#define scsi_sglist(cmd) ((struct scatterlist *)(cmd)->request_buffer)
#endif
#ifndef scsi_for_each_sg
	#define scsi_for_each_sg(cmd, sg, nseg, __i)				\
	 for (__i = 0, sg = scsi_sglist(cmd); __i < (nseg); __i++, (sg)++)

#endif

#ifndef bool
	#define bool int
#endif

#ifndef false
	#define false 0
#endif
#ifndef true
	#define true 1
#endif

#define osm_vmaloc	vmalloc
#define osm_vfree		vfree
/*
************************************************************
**  
************************************************************
*/
#ifdef CONFIG_64BIT
	#define __KCONF_64BIT__
#endif /* CONFIG_64BIT */
/*
************************************************************
**  
************************************************************
*/
#if defined(__KCONF_64BIT__)||defined(_64_SYS_)
	#define _SUPPORT_64_BIT
#else
	#ifdef _SUPPORT_64_BIT
		#error Error 64_BIT CPU Macro
	#endif
#endif
#if defined(_SUPPORT_64_BIT)
	typedef unsigned long long	PTR_INTEGER;
#else
	typedef unsigned long	PTR_INTEGER;
#endif
/*
************************************************************
**
************************************************************
*/
#if defined(_SUPPORT_64_BIT)
	#define PtrToNum(p)		(unsigned long long)(void *)(p)
	#define NumToPtr(ul)		(void *)((unsigned long long)ul)
#else
	#define PtrToNum(p)		(unsigned int)(unsigned char *)(p)
	#define NumToPtr(ul)		(void *)((unsigned long)ul)
#endif

#define osm_roundup(x, y) (		\
{					\
	const typeof(y) __y = y;		\
	(((x) + (__y - 1)) / __y) * __y;	\
}					\
)
/*
**************************************************************************************
**************************************************************************************
*/
#if !defined(sg_page) && (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 24))
	#define sg_page(sg)		sg->page
#endif
/*
**************************************************************************************
**************************************************************************************
*/
struct osm_list_head {
	struct osm_list_head *next, *prev;
};
#define osm_list_entry(ptr, type, member)				\
	container_of(ptr, type, member)

#define osm_list_first_entry(ptr, type, member)			\
	list_entry((ptr)->next, type, member)

#define list_for_each(pos, head)					\
	for (pos = (head)->next; prefetch(pos->next), pos != (head);	\
		pos = pos->next)

#define __osm_list_for_each(pos, head)				\
	for (pos = (head)->next; pos != (head); pos = pos->next)

#define osm_list_for_each_entry(pos, head, member)			\
	for (pos = osm_list_entry((head)->next, typeof(*pos), member);	\
	     prefetch(pos->member.next), &pos->member != (head);	\
	     pos = osm_list_entry(pos->member.next, typeof(*pos), member))
/*
**************************************************************************************
**************************************************************************************
*/
#define osm_msleep(x)		msleep(x)
#define osm_mdelay(x)		mdelay(x)
#define osm_udelay(x)		udelay(x)

/*
**************************************************************************************
**************************************************************************************
*/
#define osm_printk(format, arg...)		\
	do {				\
		printk(format, ##arg);	\
		mdelay(6);		\
	} while (0)
#define _ASSERT()  do{ osm_printk("ASSERT: function:%s, line:%d\n", __FUNCTION__, __LINE__);}while(0)
#define OSM_INIT_WORK(_work, _func, _dev...)		INIT_WORK(_work, _func, ##_dev)
/*
**************************************************************************************
**************************************************************************************
*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 20)
	extern void arcsas_add_scsi_device_wqfn(struct work_struct *work);
	extern void arcsas_do_smp_phy_control_reset_wqfn(struct work_struct *work);
	extern void arcsas_do_smp_state_machine_wqfn(struct work_struct *work);
	extern void arcsas_remove_scsi_device_wqfn(struct work_struct *work);

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	extern void arcsas_add_scsi_device_wqfn(void *acb);
	extern void arcsas_do_smp_phy_control_reset_wqfn(void *acb);
	extern void arcsas_do_smp_state_machine_wqfn(void *acb);
	extern void arcsas_remove_scsi_device_wqfn(void *acb);
#endif
extern unsigned long get_jiffies(void);
extern void osm_atomic_inc(atomic_t *v);	
extern int osm_atomic_read(const atomic_t *v);
extern void osm_atomic_dec(atomic_t *v);
extern void osm_atomic_set(atomic_t *addr, int v);
extern int osm_pci_read_config_dword(struct pci_dev *dev, int offset, u32 *ptr);
extern int osm_pci_read_config_word(struct pci_dev *dev, int offset, u16 *ptr);
extern int osm_pci_read_config_byte(struct pci_dev *dev, int offset, u8 *ptr);
extern int osm_pci_write_config_dword(struct pci_dev *dev, int offset, u32 val);	
extern int osm_pci_write_config_word(struct pci_dev *dev, int offset, u16 val);
extern int osm_pci_write_config_byte(struct pci_dev *dev, int offset, u8 val);
extern void osm_memset(void *buf, int  value, size_t len);
extern void osm_memcpy(void *dest, void *source, size_t len);
extern int osm_memcmp(void *buf0, void *buf1, size_t len);
extern int osm_schedule_work(struct work_struct *work);			
extern void arcsas_spin_lock_init(spinlock_t *lock);
extern void arcsas_spin_lock_irqsave(spinlock_t *lock, unsigned long *flags);
extern void arcsas_spin_unlock_irqrestore(spinlock_t *lock, unsigned long *flags);
extern void arcsas_spin_lock(spinlock_t *lock);
extern void arcsas_spin_unlock(spinlock_t *lock);
extern struct scsi_device *__osm_scsi_device_lookup(struct Scsi_Host *shost, uint channel, uint id, uint lun);
extern struct scsi_device *osm_scsi_device_lookup(struct Scsi_Host *shost, uint channel, uint id, uint lun);
extern void osm_scsi_device_put(struct scsi_device *sdev);
extern void osm_writel(u32 val, void *addr);
extern u32 osm_readl(void *addr);
extern void osm_outl(u32 val, unsigned long addr);
extern u32 osm_inl(unsigned long addr);
extern void arcsas_touch_watchdog(void);
extern u32 arcsas_display_time(void);

