/*
*************************************************************************************************
**	O.S		: Linux
**	FILE NAME	: arcsas.c
**	BY		: Nick Cheng
**	Description	: SCSI Device Driver for Areca Host Bus Adapter
*******************************************************************************
** Copyright (C) 2002 - 2005, Areca Technology Corporation All rights reserved.
**
**	Web site: www.areca.com.tw
**	E-mail: support@areca.com.tw
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU General Public License for more details.
*******************************************************************************
** Redistribution and use in source and binary forms,with or without
** modification,are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**    notice,this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice,this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
** 3. The name of the author may not be used to endorse or promote products
**    derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
** IMPLIED WARRANTIES,INCLUDING,BUT NOT LIMITED TO,THE IMPLIED WARRANTIES
** OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
** IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,INDIRECT,
** INCIDENTAL,SPECIAL,EXEMPLARY,OR CONSEQUENTIAL DAMAGES(INCLUDING,BUT
** NOT LIMITED TO,PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
** DATA,OR PROFITS; OR BUSINESS INTERRUPTION)HOWEVER CAUSED AND ON ANY
** THEORY OF LIABILITY,WHETHER IN CONTRACT,STRICT LIABILITY,OR TORT
**(INCLUDING NEGLIGENCE OR OTHERWISE)ARISING IN ANY WAY OUT OF THE USE OF
** THIS SOFTWARE,EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*********************************************************************************
** History
**
** Beta Version		DATE		NAME		Description
*************************************************************************************************
*/
#include "arc13xx.h"
/*
*******************************************************************************
**
*******************************************************************************
*/
#if ARCSAS_SUPPORT_IOCTL
	static bool arcsas_core_ccb_get_scsi_device_info(PCore_Extension pCore, struct CommandControlBlock *pCCB);
	static bool arcsas_core_ccb_get_expander_info(PCore_Extension pCore, struct CommandControlBlock *pCCB);
	static bool arcsas_core_ccb_get_pm_info(PCore_Extension pCore, struct CommandControlBlock *pCCB);
	static bool arcsas_core_ccb_set_disk_config(PCore_Extension pCore, struct CommandControlBlock *pCCB);
	static bool arcsas_core_ccb_bsl_dump(PCore_Extension pCore, struct CommandControlBlock *pCCB);
	static bool arcsas_core_ccb_get_disk_smart_status(PCore_Extension pCore, struct CommandControlBlock *pCCB);
	static void arcsas_expander_info_data_update(PCore_Extension pCore, PDomain_Expander pExp, PExpander_Info pExpInfo, struct CommandControlBlock *pCCB);
	extern void arcsas_handle_virtual_ses_device(PACB pACB, struct scsi_cmnd *pscsi_cmnd);
#endif
extern void arcsas_remove_device(PDomain_Device pDevice);
extern bool arcsas_device_babies_lun_check(PDomain_Device pDevice);
extern void arcsas_send_ses_diagnostic_for_all_enclosure(PACB pACB);
extern bool arcsas_core_sgpio_led_activated(PDomain_Device pDevice, unsigned char flag);
#if ARCSAS_SUPPORT_IOCTL || ARCSAS_SUPPORT_SMART
	static void arcsas_hba_smart_on_ccb_callback(void *pCore, PCCB pCCB);
#endif
static bool arcsas_core_initialize(struct AdapterControlBlock *pACB);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 37)
	static int arcsas_queuecommand(struct Scsi_Host *h, struct scsi_cmnd *cmd);
#else
	static int arcsas_queuecommand(struct scsi_cmnd *cmd, void (* done)(struct scsi_cmnd *));
#endif
#if defined(RHEL_RELEASE_CODE) && LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 31)
	static int arcsas_change_queue_depth(struct scsi_device *sdev, int queue_depth, int reason);
#elif LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 32)
	static int arcsas_change_queue_depth(struct scsi_device *sdev, int queue_depth, int reason);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 11)
	static int arcsas_change_queue_depth(struct scsi_device *sdev, int queue_depth);
#else
	static ssize_t arcsas_change_queue_depth(struct device *dev, const char *buf, size_t count);
#endif
static int arcsas_pci_probe(struct pci_dev *pcidev, const struct pci_device_id *id);
static void arcsas_pci_remove(struct pci_dev *pcidev);
void arcsas_delete_device_timer(struct AdapterControlBlock *pACB, PDomain_Device pDevice);
void arcsas_hba_smart_cancel_timer(PACB pACB);
static unsigned long arcsas_build_os_ccb(struct AdapterControlBlock *pACB, struct scsi_cmnd *pscsi_cmnd, struct _Domain_Device **pDevice);
void arcsas_done_os_ccb_callback(void *pACB, struct CommandControlBlock *pCCB);
struct CommandControlBlock *arcsas_core_get_coreccb_from_list(PCore_Extension pCore);
void arcsas_core_free_coreccb_to_pool(PCore_Extension pCore, struct CommandControlBlock *pCCB);
void arcsas_os_ccb_done(struct AdapterControlBlock *pACB, struct CommandControlBlock *pCCB);
static int arcsas_os_eh_abort_handler(struct scsi_cmnd *pscsicmd);
#if 0
static int arcsas_os_eh_bus_reset_handler(struct scsi_cmnd *pscsicmd);
static int arcsas_os_eh_dev_reset_handler(struct scsi_cmnd *pscsicmd);
#endif
/* scsi-mid layer global parmeter is max_report_luns, which is 511 */
static int max_lun = 1024;
module_param(max_lun, int, 0);
MODULE_PARM_DESC(max_lun, " max lun, default = 1024 ");
/*
*********************************************************************************
*********************************************************************************
*/
static struct pci_device_id arcsas_pci_id_table[] __devinitdata = {
	{PCI_DEVICE(PCI_VENDOR_ID_ARECA, DEVICE_ID_1300)},
	{PCI_DEVICE(PCI_VENDOR_ID_ARECA, DEVICE_ID_1320)},
	{} /* Terminating entry */
};
/*
*********************************************************************************
*********************************************************************************
*/
static struct pci_driver arcsas_pci_driver = {
	.name		= "arcsas",
	.id_table		= arcsas_pci_id_table,
	.probe		= arcsas_pci_probe,
	.remove		= arcsas_pci_remove,
	//.shutdown	= arcsas_shutdown,
};
/*
*********************************************************************************
*********************************************************************************
*/
MODULE_AUTHOR("Nick Cheng <support@areca.com.tw>");
MODULE_DESCRIPTION("Areca SAS Host Bus Adapter");
MODULE_VERSION(ARCSAS_DRIVER_VERSION);
#ifdef MODULE_LICENSE
	MODULE_LICENSE("GPL");
#endif
MODULE_DEVICE_TABLE(pci, arcsas_pci_id_table);
#if 1
/*
*********************************************************************************
* reset nmi watchdog before delay
*********************************************************************************
*/
void arcsas_touch_watchdog(void)
{
	#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 13)
		touch_nmi_watchdog();
	#else
		touch_all_softlockup_watchdogs();
	#endif /* LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 13) */
}
#endif
/*
**************************************************************************************
**************************************************************************************
*/
#define _WARN(condition, format...) ({			\
	int __ret_warn_on = !!(condition);			\
	unlikely(__ret_warn_on);				\
})
#define _WARN_ONCE(condition, format...)({			\
	static int __warned;					\
	int __ret_warn_once = !!(condition);			\
							\
	if (unlikely(__ret_warn_once))			\
		if (WARN(!__warned, format)) 		\
			__warned = 1;			\
	unlikely(__ret_warn_once);				\
})
/*
**************************************************************************************
**************************************************************************************
*/
u32 arcsas_display_time(void)
{
	static u32 utc_time = 0, local_time = 0;
	struct timeval tv;
	do_gettimeofday(&tv);
	local_time = (u32)((u32) tv.tv_sec - (sys_tz.tz_minuteswest * 60));
	return local_time;
}
/*
**************************************************************************************
**************************************************************************************
*/
/* register read write: memory io */
void osm_writel(u32 val, void *addr)	{writel(val, addr);}
u32 osm_readl(void *addr)		{return readl(addr);}
/* register read write: port io */
void osm_outl(u32 val, unsigned long addr)	{outl(val, addr);}
u32 osm_inl(unsigned long addr)		{return inl(addr);}
/*
**************************************************************************************
**************************************************************************************
*/
inline void osm_memset(void *buf, int  value, size_t len)	{memset(buf, value, len);}
inline void osm_memcpy(void *dest, void *source, size_t len)	{memcpy(dest, source, len);}
inline int osm_memcmp(void *buf0, void *buf1, size_t len)	{return memcmp(buf0, buf1, len);}
/*
**************************************************************************************
**************************************************************************************
*/
int osm_schedule_work(struct work_struct *work)	{ return schedule_work(work);}			
/*
**************************************************************************************
**************************************************************************************
*/
inline void osm_atomic_inc(atomic_t *v)		{atomic_inc(v);}	
inline int osm_atomic_read(const atomic_t *v)	{return atomic_read(v);}
inline void osm_atomic_dec(atomic_t *v)		{atomic_dec(v);}
inline void osm_atomic_set(atomic_t *addr, int v)	{atomic_set(addr, v);}
/*
**************************************************************************************
**************************************************************************************
*/
inline int osm_pci_read_config_dword(struct pci_dev *dev, int offset, u32 *ptr)	{return pci_read_config_dword(dev, offset, ptr);}
inline int osm_pci_read_config_word(struct pci_dev *dev, int offset, u16 *ptr)	{return pci_read_config_word(dev, offset, ptr);}
inline int osm_pci_read_config_byte(struct pci_dev *dev, int offset, u8 *ptr)		{return pci_read_config_byte(dev, offset, ptr);}
inline int osm_pci_write_config_dword(struct pci_dev *dev, int offset, u32 val)	{return pci_write_config_dword(dev, offset, val);}	
inline int osm_pci_write_config_word(struct pci_dev *dev, int offset, u16 val)	{return pci_write_config_word(dev, offset, val);}
inline int osm_pci_write_config_byte(struct pci_dev *dev, int offset, u8 val)		{return pci_write_config_byte(dev, offset, val);}
/*
**************************************************************************************
**************************************************************************************
*/
inline void arcsas_spin_lock_init(spinlock_t *lock)
{
	spin_lock_init(lock);
}
inline void arcsas_spin_lock_irqsave(spinlock_t *lock, unsigned long *flags)
{
	spin_lock_irqsave(lock, *flags);
}
inline void arcsas_spin_unlock_irqrestore(spinlock_t *lock, unsigned long *flags)
{
	spin_unlock_irqrestore(lock, *flags);
}
inline void arcsas_spin_lock(spinlock_t *lock)
{
	spin_lock(lock);
}
inline void arcsas_spin_unlock(spinlock_t *lock)
{
	spin_unlock(lock);
}
/*
**************************************************************************************
**************************************************************************************
*/
inline struct scsi_device *__osm_scsi_device_lookup(struct Scsi_Host *shost, uint channel, uint id, uint lun)
{
	return __scsi_device_lookup(shost, channel, id, lun);
}
/*
**************************************************************************************
**************************************************************************************
*/
inline struct scsi_device *osm_scsi_device_lookup(struct Scsi_Host *shost, uint channel, uint id, uint lun)
{
	return __scsi_device_lookup(shost, channel, id, lun);
}
/*
**************************************************************************************
**************************************************************************************
*/
inline void osm_scsi_device_put(struct scsi_device *sdev)
{
	scsi_device_put(sdev);
}
/*
**************************************************************************************
**************************************************************************************
*/
inline void __osm_list_del(struct osm_list_head * prev, struct osm_list_head * next)
{
	next->prev = prev;
	prev->next = next;
}

void __osm_list_add(struct osm_list_head *new, struct osm_list_head *prev, struct osm_list_head *next)
{
#if 1
	_WARN(next->prev != prev,
		"osm_list_add corruption. next->prev should be "
		"prev (%p), but was %p. (next=%p).\n",
		prev, next->prev, next);
	_WARN(prev->next != next,
		"osm_list_add corruption. prev->next should be "
		"next (%p), but was %p. (prev=%p).\n",
		next, prev->next, prev);
#endif
	next->prev = new;
	new->next = next;
	new->prev = prev;
	prev->next = new;
}

void osm_list_del(struct osm_list_head *entry)
{
#if 1
	_WARN(entry->next == LIST_POISON1,
		"osm_list_del corruption, next is LIST_POISON1 (%p)\n",
		LIST_POISON1);
	_WARN(entry->next != LIST_POISON1 && entry->prev == LIST_POISON2,
		"osm_list_del corruption, prev is LIST_POISON2 (%p)\n",
		LIST_POISON2);
	_WARN(entry->prev->next != entry,
		"osm_list_del corruption. prev->next should be %p, "
		"but was %p\n", entry, entry->prev->next);
	_WARN(entry->next->prev != entry,
		"osm_list_del corruption. next->prev should be %p, "
		"but was %p\n", entry, entry->next->prev);
#endif
	__osm_list_del(entry->prev, entry->next);
	entry->next = LIST_POISON1;
	entry->prev = LIST_POISON2;
}

inline void osm_list_add_tail(struct osm_list_head *new, struct osm_list_head *head)
{
	__osm_list_add(new, head->prev, head);
}

inline void OSM_INIT_LIST_HEAD(struct osm_list_head *list)
{
	list->next = list;
	list->prev = list;
}

inline void osm_list_del_init(struct osm_list_head *entry)
{
	__osm_list_del(entry->prev, entry->next);
	OSM_INIT_LIST_HEAD(entry);
}

inline void osm_list_add(struct osm_list_head *new, struct osm_list_head *head)
{
	__osm_list_add(new, head, head->next);
}

inline void osm_list_move(struct osm_list_head *list, struct osm_list_head *head)
{
	__osm_list_del(list->prev, list->next);
	osm_list_add(list, head);
}

inline void osm_list_move_tail(struct osm_list_head *list, struct osm_list_head *head)
{
	__osm_list_del(list->prev, list->next);
	osm_list_add_tail(list, head);
}

inline int osm_list_empty(const struct osm_list_head *head)
{
	return head->next == head;
}
unsigned long get_jiffies(void)
{
	return jiffies;
}
void *osm_kmalloc(size_t size, gfp_t flags){	return kmalloc(size, flags);}
void osm_kfree(const void *x){	return kfree(x);}
/*
***********************************************************************************
***********************************************************************************
*/
#ifndef scsilun_to_int
int scsilun_to_int(struct scsi_lun *scsilun)
{
	int i;
	unsigned int lun;

	lun = 0;
	for (i = 0; i < sizeof(lun); i += 2)
		lun = lun | (((scsilun->scsi_lun[i] << 8) |
			      scsilun->scsi_lun[i + 1]) << (i * 8));
	return lun;
}
#endif
/*
***********************************************************************************
***********************************************************************************
*/
#define SAS_MAX_QD 64
/*
***********************************************************************************
**   initialize  timer request
***********************************************************************************
*/
inline void arcsas_init_timer(struct timer_list *timer)
{
	init_timer(timer);
	timer->expires = 0;
	timer->function  = NULL;
	timer->data = 0;
	return;
}
/*
************************************************************************************
** This is identical to Timer_AddRequest() in Odin.
************************************************************************************
*/
unsigned char arcsas_run_timer(struct timer_list *timer, unsigned long timeout, Timer_CBFunc function, unsigned long timerData)
{
	u64 jif;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	if (timer_pending(timer))
		del_timer(timer);
	timer->function = function;
	timer->data = timerData;
	jif = (u64) (timeout * HZ);
	mod_timer(timer, jiffies + jif);
	return RUNNING;
}
/*
************************************************************************
**
************************************************************************
*/
void arcsas_hba_smart_cancel_timer(PACB pACB)
{
	arcsas_spin_lock(&pACB->SMART_timer_lock);
	if (osm_atomic_read(&pACB->SMART_timer_status) == RUNNING) {
		del_timer(&pACB->SMART_timer_request);
		osm_atomic_set(&pACB->SMART_timer_status, STOP_RUNNING);
	}
	arcsas_spin_unlock(&pACB->SMART_timer_lock);
	return;
}
/*
*******************************************************************************
**arcsas_device_cancel_timer
*******************************************************************************
*/
void arcsas_delete_device_timer(struct AdapterControlBlock *pACB, PDomain_Device pDevice)
{
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	arcsas_spin_lock(&pDevice->timer_lock);
	if (osm_atomic_read(&pDevice->Timer_Status) == RUNNING) {
		del_timer(&pDevice->timer_request);
		osm_atomic_set(&pDevice->Timer_Status, STOP_RUNNING);
	}
	arcsas_spin_unlock(&pDevice->timer_lock);
	return;
}
/*
*******************************************************************************
**arcsas_pm_cancel_timer
*******************************************************************************
*/
void arcsas_delete_pm_timer(struct AdapterControlBlock *pACB, PDomain_PM pPM)
{
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	arcsas_spin_lock(&pPM->timer_lock);
	if (pPM->Timer_Status == RUNNING) {
		del_timer(&pPM->timer_request);
		pPM->Timer_Status = STOP_RUNNING;
	}
	arcsas_spin_unlock(&pPM->timer_lock);
	return;
}
/*
*******************************************************************************
*******************************************************************************
*/
void arcsas_map_sg_to_buffer(struct CommandControlBlock *pCCB)
{
	struct scsi_cmnd *pscsi_cmnd;
	struct scatterlist *sg;
	unsigned long i;
	void * virt_address;
	unsigned long flags;

	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	pscsi_cmnd = (struct scsi_cmnd *)pCCB->Org_Req;
	sg = (struct scatterlist *)arcsas_rq_bf(pscsi_cmnd);
	if (arcsas_use_sg(pscsi_cmnd)) {
		pCCB->Data_Buffer = kmalloc(sg->length, GFP_ATOMIC);
		pCCB->Data_Transfer_Length = 0;
		if (pCCB->Data_Buffer) {
			for (i = 0; i < arcsas_use_sg(pscsi_cmnd); i++, sg++) {
				local_irq_save(flags);
				virt_address = arcsas_map_sg_page(sg);
				#if (ARCSAS_DBG_CONTEXT  || ARCSAS_DBG_OS_CCB_CONTEXT)
					osm_printk("%s: virt_address = 0x%p\n", __func__, virt_address);
				#endif
				memcpy(pCCB->Data_Buffer + pCCB->Data_Transfer_Length , virt_address + sg->offset, sg->length);
				kunmap_atomic(virt_address, KM_IRQ0);
				pCCB->Data_Transfer_Length += sg->length;
				local_irq_restore(flags);
			}
		} else {
			_ASSERT();
		}
		#if ARCSAS_DBG_CONTEXT
			osm_printk("%s: Data_Transfer_Length = 0x%d.\n", __func__, pCCB->Data_Transfer_Length);
		#endif
	} else {
		#if (ARCSAS_DBG_CONTEXT  || ARCSAS_DBG_OS_CCB_CONTEXT)
			osm_printk("%s: no use sg\n", __func__);
		#endif
		pCCB->Data_Buffer = arcsas_rq_bf(pscsi_cmnd);
	}
}
/*
*******************************************************************************
*******************************************************************************
*/
void arcsas_unmap_sg_to_buffer(struct CommandControlBlock *pCCB)
{
	struct scsi_cmnd *pscsi_cmnd = NULL;
	struct scatterlist *sg = NULL;
	unsigned long i;
	void *virt_address;
	unsigned long flags;

	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	pscsi_cmnd = (struct scsi_cmnd *)pCCB->Org_Req;
	sg = (struct scatterlist *)arcsas_rq_bf(pscsi_cmnd);
	if (arcsas_use_sg(pscsi_cmnd)) {
		pCCB->Data_Transfer_Length = 0;
		for (i = 0; i < arcsas_use_sg(pscsi_cmnd); i++, sg++) {
			local_irq_save(flags);
			virt_address = arcsas_map_sg_page(sg);
			#if (ARCSAS_DBG_CONTEXT  || ARCSAS_DBG_OS_CCB_CONTEXT)
				osm_printk("%s: virt_address = 0x%p\n", __func__, virt_address);
			#endif
			memcpy(virt_address + sg->offset, pCCB->Data_Buffer + pCCB->Data_Transfer_Length, sg->length);
			#if (ARCSAS_DBG_CONTEXT  || ARCSAS_DBG_OS_CCB_CONTEXT)
				osm_printk("%s: kunmap_atomic\n", __func__);
			#endif
			kunmap_atomic(virt_address , KM_IRQ0);
			pCCB->Data_Transfer_Length += sg->length;
			local_irq_restore(flags);
		}
		kfree(pCCB->Data_Buffer);
		pCCB->Data_Buffer = arcsas_rq_bf(pscsi_cmnd);
	}
}
/*
*******************************************************************************
**
*******************************************************************************
*/
/* Release the buffer and set the pointer to NULL. */
void arcsas_core_free_sas_scratch_buffer_to_list(PCore_Extension pCore, PSAS_Scratch_Buffer pSASSB)
{
	//unsigned long flags;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	if (pSASSB) {
		arcsas_spin_lock(&pCore->sassb_list_lock);
		osm_list_add_tail(&pSASSB->Queue_Pointer, &pCore->SAS_Scratch_List);
		arcsas_spin_unlock(&pCore->sassb_list_lock);
		pSASSB = NULL;
	}
}
/*
*******************************************************************************
**
*******************************************************************************
*/
PSAS_Scratch_Buffer arcsas_core_get_sas_scratch_buffer_from_list(PCore_Extension pCore)
{
	PSAS_Scratch_Buffer pSASSB;
	//unsigned long flags;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	arcsas_spin_lock(&pCore->sassb_list_lock);
	if (!osm_list_empty(&pCore->SAS_Scratch_List)) {
		pSASSB = (PSAS_Scratch_Buffer)osm_list_first_entry(&pCore->SAS_Scratch_List, struct _SAS_Scratch_Buffer, Queue_Pointer);
		osm_list_del_init(&pSASSB->Queue_Pointer);
		arcsas_spin_unlock(&pCore->sassb_list_lock);
		return pSASSB;
	} else {
		arcsas_spin_unlock(&pCore->sassb_list_lock);
		return NULL;
	}
}
/*
**********************************************************************
**
**********************************************************************
*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 20)
	void arcsas_add_scsi_device_wqfn(struct work_struct *work)
	{
		struct _Domain_Device *pDevice;
		int rc;
		struct AdapterControlBlock *pacb;
		struct _Domain_Port *pPort;

		#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_PLUG_UNPLUG)
			osm_printk("%s:\n", __func__);
		#endif
		pDevice = container_of(work, struct _Domain_Device, arcsas_add_scsi_device);
		pacb = pDevice->pCore->pACB;
		pPort = pDevice->pPort;
		rc = scsi_add_device(pacb->host, 0, DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex), DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex));
		if (rc) {
			osm_printk("Failed to add SCSI device, id:lun = %d:%d\n", DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex), DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex));
		} else {
			osm_printk("SCSI device id:lun = %d:%d, added @port %d\n", DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex), DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex), pDevice->pPort->Port_Index);
		}
		return;
	}
	/*
	**********************************************************************
	arcsas_core_notify_device_hotplug(pCore,false,EntryIndex);  scsi id lun EVENT_DEVICE_REMOVAL
	**********************************************************************
	*/
	void arcsas_remove_scsi_device_wqfn(struct work_struct *work)
	{
		struct _Domain_Device *pDevice;
		struct scsi_device *psdev;
		PACB pACB = acb_addr;
		unsigned long flags;
		#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_PLUG_UNPLUG)
			osm_printk("%s:\n", __func__);
		#endif
		spin_lock_irqsave(&pACB->host_lock, flags);
		pDevice = container_of(work, struct _Domain_Device, arcsas_remove_scsi_device);
		psdev = pDevice->sdev;
		if (psdev != NULL) {
			if (IS_SGPIO(pDevice)) {
				arcsas_core_sgpio_led_activated(pDevice, SES_LED_FLAG_ALL_OFF);
			} else if (IS_SSP(pDevice) || IS_STP(pDevice)) {
				#if ARCSAS_SUPPORT_SES
					arcsas_core_ses_led_activated(pDevice, SES_LED_FLAG_ALL_OFF);
				#endif
			}
			scsi_remove_device(psdev);
			osm_printk("SCSI device(0x%p), id:lun = %d:%d, removed @port %d\n", psdev, DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex), DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex), pDevice->pPort->Port_Index);
		}
		spin_unlock_irqrestore(&pACB->host_lock, flags);
		return;
	}
	/*
	*******************************************************************************
	*******************************************************************************
	*/
	void arcsas_do_smp_state_machine_wqfn(struct work_struct *work)
	{
		struct _Discover_Extension *pDiscovery;
		PACB pACB;
		PDISCOVERY_ELEMENT pDiscovery_element;
		#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_DISCOVERY)
			osm_printk("%s:\n", __func__);
		#endif
		pDiscovery = container_of(work, struct _Discover_Extension, arcsas_do_smp_state_machine);
		pACB = pDiscovery->Target->pPort->pCore->pACB;
		pDiscovery_element = pACB->pDiscovery_element;
		arcsas_discovery_smp_state_machine(pDiscovery_element->pDiscovery, pDiscovery_element->pPort, pDiscovery_element->nTgt, pDiscovery_element->CurrentState, pDiscovery_element->context, pDiscovery_element->pBuffer);
		pACB->discoveryQ_count--;
		return;
	}
	/*
	**********************************************************************
	**
	**********************************************************************
	*/
	void arcsas_do_smp_phy_control_reset_wqfn(struct work_struct *work)
	{
		struct _Discover_Extension *pDiscovery;
		PACB pACB;
		PDISCOVERY_ELEMENT pDiscovery_element;
		PDomain_Port pPort;
		unsigned char nTgt;
		unsigned char PhyId;
		unsigned char CurrentState;
		#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_DISCOVERY)
			osm_printk("%s:\n", __func__);
		#endif
		pDiscovery = container_of(work, struct _Discover_Extension, arcsas_do_smp_phy_control_reset);
		pACB = pDiscovery->Target->pPort->pCore->pACB;
		pDiscovery_element = pACB->pDiscovery_element;
		pPort = pDiscovery_element->pPort;
		nTgt = pDiscovery_element->nTgt;
		PhyId = pDiscovery_element->PhyId;
		CurrentState = pDiscovery_element->CurrentState;
		pDiscovery->numOfDiscoveryReq++;
		if (arcsas_post_smp_issue_phy_control_reset(pDiscovery, pPort, nTgt, PhyId) == STATUS_REQUEST_OK) {/*issue PHY_CONTROL SMP function ,let sata oob done*/
			osm_printk("issue phy control .. reset ....................\n");
			pACB->discoveryQ_count--;
		} else {
			osm_printk("arcsas_post_smp_issue_phy_control_reset, we got a queue full....\n");
			osm_printk("issue phy control .. reset ....not ok................\n");
			pACB->discoveryQ_count--;
			pDiscovery->numOfDiscoveryReq--;
			arcsas_construct_smp_state_element(pDiscovery, pPort, nTgt, PhyId, CurrentState, 0, NULL);
			schedule_work(&pDiscovery->arcsas_do_smp_phy_control_reset);
		}
		osm_printk("%s: line %d numOfDiscoveryReq = %d\n", __func__, __LINE__, pDiscovery->numOfDiscoveryReq);
		return;
	}
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	void arcsas_add_scsi_device_wqfn(void *ptr)
	{
		struct _Domain_Device *pDevice;
		int rc;
		struct AdapterControlBlock *pacb;
		struct _Domain_Port *pPort;

		#if ARCSAS_DBG_FUNCTION
			osm_printk("%s:\n", __func__);
		#endif
		pDevice = (struct _Domain_Device *)ptr;
		pacb = pDevice->pCore->pACB;
		pPort = pDevice->pPort;
		rc = scsi_add_device(pacb->host, 0, DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex), DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex));
		if (rc) {
			osm_printk("Failed to add SCSI device, id:lun = %d:%d\n", DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex), DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex));
		} else {
			osm_printk("SCSI device id:lun = %d:%d, added @port %d\n", DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex), DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex), pDevice->pPort->Port_Index);
		}
		return;
	}
	/*
	*******************************************************************************
	arcsas_core_notify_device_hotplug(pCore,false,EntryIndex); scsi id lun EVENT_DEVICE_REMOVAL
	*******************************************************************************
	*/
	void arcsas_remove_scsi_device_wqfn(void *ptr)
	{
		struct _Domain_Device *pDevice;
		struct scsi_device *psdev;
		PCore_Extension pCore;
		PACB pACB = acb_addr;
		unsigned long flags;
		#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_PLUG_UNPLUG)
			osm_printk("%s:\n", __func__);
		#endif
		spin_lock_irqsave(&pACB->host_lock, flags);
		pDevice = (struct _Domain_Device *)ptr;
		psdev = pDevice->sdev;
		if (psdev != NULL) {
			if (IS_SGPIO(pDevice)) {
				arcsas_core_sgpio_led_activated(pDevice, SES_LED_FLAG_ALL_OFF);
			} else if (IS_SSP(pDevice) || IS_STP(pDevice)) {
				#if ARCSAS_SUPPORT_SES
					arcsas_core_ses_led_activated(pDevice, SES_LED_FLAG_ALL_OFF);
				#endif
			}
			scsi_remove_device(psdev);
			osm_printk("SCSI device(0x%p), id:lun = %d:%d, removed @port %d\n", psdev, DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex), DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex), pDevice->pPort->Port_Index);
		}
		spin_unlock_irqrestore(&pACB->host_lock, flags);
		return;
	}
	/*
	*******************************************************************************
	*******************************************************************************
	*/
	void arcsas_do_smp_state_machine_wqfn(void *ptr)
	{
		struct _Discover_Extension *pDiscovery;
		PDISCOVERY_ELEMENT pDiscovery_element;
		PACB pACB;
		#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_DISCOVERY)
			osm_printk("%s:\n", __func__);
		#endif
		pDiscovery = (struct _Discover_Extension *)ptr;
		pACB = pDiscovery->Target->pPort->pCore->pACB;
		pDiscovery_element = pACB->pDiscovery_element;
		arcsas_discovery_smp_state_machine(pDiscovery_element->pDiscovery, pDiscovery_element->pPort, pDiscovery_element->nTgt, pDiscovery_element->CurrentState, pDiscovery_element->context, pDiscovery_element->pBuffer);
		pACB->discoveryQ_count--;
		return;
	}
	/*
	**********************************************************************
	**
	**********************************************************************
	*/
	void arcsas_do_smp_phy_control_reset_wqfn(void *ptr)
	{
		struct _Discover_Extension *pDiscovery;
		PACB pACB;
		PDISCOVERY_ELEMENT pDiscovery_element;
		PDomain_Port pPort;
		unsigned char nTgt;
		unsigned char PhyId;
		unsigned char CurrentState;
		#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_DISCOVERY)
			osm_printk("%s:\n", __func__);
		#endif
		pDiscovery = (struct _Discover_Extension *)ptr;
		pACB = pDiscovery->Target->pPort->pCore->pACB;
		pDiscovery_element = pACB->pDiscovery_element;
		pPort = pDiscovery_element->pPort;
		nTgt = pDiscovery_element->nTgt;
		PhyId = pDiscovery_element->PhyId;
		CurrentState = pDiscovery_element->CurrentState;
		pDiscovery->numOfDiscoveryReq++;
		if (arcsas_post_smp_issue_phy_control_reset(pDiscovery, pPort, nTgt, PhyId) == STATUS_REQUEST_OK) {/*issue PHY_CONTROL SMP function ,let sata oob done*/
			osm_printk("issue phy control .. reset ....................\n");
			pACB->discoveryQ_count--;
		} else {
			osm_printk("arcsas_post_smp_issue_phy_control_reset, we got a queue full....\n");
			osm_printk("issue phy control .. reset ....not ok................\n");
			pACB->discoveryQ_count--;
			pDiscovery->numOfDiscoveryReq--;
			arcsas_construct_smp_state_element(pDiscovery, pPort, nTgt, PhyId, CurrentState, 0, NULL);
			schedule_work(&pDiscovery->arcsas_do_smp_phy_control_reset);
		}
		osm_printk("%s: line %d numOfDiscoveryReq = %d\n", __func__, __LINE__, pDiscovery->numOfDiscoveryReq);
		return;
	}

#endif

/*
*******************************************************************************
**	Completion Callback: arcsas_os_sas_core_ccb_callback
*******************************************************************************
*/
void arcsas_os_sas_send_report_lun_taskccb(PDomain_Device pDevice, PCCB pOSCCB)
{
	struct _Core_Extension *pCore;
	PCCB pCCB;
	PSAS_Scratch_Buffer pSASSB;
	PARCSAS_SG_Table pSGTable;
	unsigned int length;

	#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_OS_CCB)
		osm_printk("%s:\n", __func__);
	#endif
	pCore = pDevice->pCore;
	pCCB = arcsas_core_get_coreccb_from_list(pCore);
	if (pCCB == NULL) {
		/* no internal requests available - wait for now */
		pDevice->Is_Waiting = true; 
		return;
	}
	pSASSB = arcsas_core_get_sas_scratch_buffer_from_list(pCore);
	if (pSASSB == NULL) {
		pDevice->Is_Waiting = true;
		arcsas_core_free_coreccb_to_pool(pCore, pCCB);
		#if ARCSAS_DBG_OS_CCB_CONTEXT
			osm_printk("ERROR: No more free SAS scratch buffer. Request aborted.\n");
		#endif
		return;
	}
	pSGTable = &pCCB->SG_Table;
	pCCB->Cdb[0] = SCSI_CMD_REPORT_LUN;
	osm_memset(&pCCB->Cdb[1], 0, 5);/* bytes 1 - 5: reserved,set to zero. Cdb[2]:00b ....SELECT REPORT  Peripheral device addressing method */
	length = (ARCSAS_MAX_TARGETLUN + 1) * 8;
	pCCB->Cdb[6] = (unsigned char) (length >> 24) & 0xff;/*MSB:ALLOCATION LENGTH*/
	pCCB->Cdb[7] = (unsigned char) (length >> 16) & 0xff;
	pCCB->Cdb[8] = (unsigned char) (length >> 8) & 0xff;
	pCCB->Cdb[9] = (unsigned char) length & 0xff;/*LSB:ALLOCATION LENGTH*/
	pCCB->Cdb[10] = 0;	/* reserved */
	pCCB->Cdb[11] = 0;	/* control */
	pCCB->Tag = 0xaa;
	pCCB->Org_Req = pOSCCB;
	pCCB->EntryIndex = pDevice->EntryIndex;
	#if ARCSAS_DBG_OS_CCB_CONTEXT
		osm_printk(" %s: pCCB->EntryIndex =  0x%x, pDevice =  0x%p.\n", __func__, pCCB->EntryIndex, pDevice);
	#endif
	pCCB->pDevice = pDevice;
	pCCB->Cmd_Initiator = (void *)pCore;
	pCCB->Data_Transfer_Length = 0x60;/* length need to enlarge and support 128 lun device if needed */
	pCCB->Cmd_Flag |= CMD_FLAG_DATA_IN;
	pCCB->Data_Buffer = pSASSB->Scratch_Vir;
	pCCB->Scratch_Buffer = pSASSB;
	pCCB->Completion = (void(*)(void *, struct CommandControlBlock *))arcsas_os_sas_core_ccb_callback;
	/* Make the SG table. */
	arcsas_sg_table_init(pSGTable, ARCSAS_MAX_CORE_SG_ENTRIES);
	arcsas_sg_table_append(pSGTable, pSASSB->Scratch_DMA.parts.low, pSASSB->Scratch_DMA.parts.high, 0x60);
	/* Send this internal request */
	arcsas_post_ccb(pCore, pCCB);
	return;
}
/*
*******************************************************************************
**This should be the _only_ os request exit point
*******************************************************************************
*/
void arcsas_done_os_inquiry_ccb_callback(PACB pACB, PCCB pCCB)
{
	struct scsi_cmnd *pscsi_cmd;
	PSCSI_INQ_DATA pInquiryData = NULL;
	struct scatterlist *sg;
	unsigned char *buffer;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	if ((DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pCCB->EntryIndex) == 0) && (pCCB->CCB_Status != CCB_STATUS_SUCCESS)) {
		pscsi_cmd = (struct scsi_cmnd *)pCCB->Org_Req;
		sg = arcsas_rq_bf(pscsi_cmd);
		buffer = kmap_atomic(sg_page(sg), KM_IRQ0) + sg->offset;
		pInquiryData = (PSCSI_INQ_DATA)buffer;
		pInquiryData->DeviceType = 0x0D;
		pInquiryData->Version = 1;
		pInquiryData->AdditionalLength = 31;
		osm_memcpy(pInquiryData->VendorId, "  Areca  ", 8);
		osm_memcpy(pInquiryData->ProductId, " DummyDevice    ", 16);
		kunmap_atomic(buffer, KM_IRQ0);
		pCCB->CCB_Status = CCB_STATUS_SUCCESS;		
	}
	arcsas_os_ccb_done(pACB, pCCB);
	return;
}
/*
*******************************************************************************
**
*******************************************************************************
*/
PSMP_Scratch_Buffer arcsas_core_get_smp_scratch_from_list(PCore_Extension pCore)
{
	PSMP_Scratch_Buffer pSMPSB;
	//unsigned long flags;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	arcsas_spin_lock(&pCore->smpsb_list_lock);
	if (!osm_list_empty(&pCore->SMP_Scratch_List)) {
		//head = arcsas_list_get_first(&pCore->SMP_Scratch_List);
		pSMPSB = (PSMP_Scratch_Buffer)osm_list_first_entry(&pCore->SMP_Scratch_List, struct _SMP_Scratch_Buffer, Queue_Pointer);
		osm_list_del_init(&pSMPSB->Queue_Pointer);
		arcsas_spin_unlock(&pCore->smpsb_list_lock);
		return pSMPSB;
	} else {
		arcsas_spin_unlock(&pCore->smpsb_list_lock);
		return NULL;
	}
}
/*
*******************************************************************************
**
*******************************************************************************
*/
void arcsas_core_free_smp_scratch_to_list(PCore_Extension pCore, PSMP_Scratch_Buffer *pSMPSB)
{
	//unsigned long flags;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	// TBD: reset SMP Request - what values need resetting?
	if (*pSMPSB) {
		arcsas_spin_lock(&pCore->smpsb_list_lock);
		#if ARCSAS_DBG_LINKEDLIST
			osm_printk("line %d \n", __LINE__);
		#endif
		osm_list_add_tail(&(*pSMPSB)->Queue_Pointer, &pCore->SMP_Scratch_List);
		arcsas_spin_unlock(&pCore->smpsb_list_lock);
	}
	return;
}
/*
*******************************************************************************
**  MAX_PRD_Buffer_COUNT = 512 (PRD: Physical Region Description)
*******************************************************************************
*/
PPRD_Buffer arcsas_core_get_prd_buffer_from_list(PCore_Extension pCore)
{
	PPRD_Buffer pPRD_Buffer;
	//unsigned long flags;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	arcsas_spin_lock(&pCore->prd_list_lock);
	if (osm_list_empty(&pCore->PRD_Buffer_List)) {
		arcsas_spin_unlock(&pCore->prd_list_lock);
		return NULL;
	}
	pPRD_Buffer = (PPRD_Buffer)osm_list_first_entry(&pCore->PRD_Buffer_List, struct _PRD_Buffer, Queue_Pointer);
	osm_list_del_init(&pPRD_Buffer->Queue_Pointer);
	arcsas_spin_unlock(&pCore->prd_list_lock);
	return pPRD_Buffer;
}
/*
*******************************************************************************
**(PRD: Physical Region Description)
*******************************************************************************
*/
void arcsas_core_free_prd_buffer_to_list(PCore_Extension pCore, PPRD_Buffer *pPRD_Buffer)
{
	//unsigned long flags;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif

	if (*pPRD_Buffer) {
		arcsas_spin_lock(&pCore->prd_list_lock);
		osm_list_add_tail(&(*pPRD_Buffer)->Queue_Pointer, &pCore->PRD_Buffer_List);
		*pPRD_Buffer = NULL;
		arcsas_spin_unlock(&pCore->prd_list_lock);
	}
	return;
}
/*
*******************************************************************************
** return id of the next active device, or ID_NOT_MAPPED if there's none
*******************************************************************************
*/
unsigned char arcsas_get_next_active_dep_entry_index(PACB pACB, unsigned short dep_entry_index)
{
	unsigned short next_dep_entry_index, temp_dep_entry_index;
	unsigned char i;

	if (dep_entry_index == ID_NOT_MAPPED) {
		for (i = 0; i < ARCSAS_MAX_REAL_DEVICE_NUMBER; i++) {
			if ((temp_dep_entry_index = Get_DevicesExpandersPMs_EntryIndex(pACB, i)) == ID_NOT_MAPPED) {
				continue;
			}
			return temp_dep_entry_index;
		}
	} else {
		for (i = 0; i < ARCSAS_MAX_REAL_DEVICE_NUMBER; i++) {
			if ((temp_dep_entry_index = Get_DevicesExpandersPMs_EntryIndex(pACB, i)) != dep_entry_index) {
				continue;
			}
			for (i = i + 1; i < ARCSAS_MAX_REAL_DEVICE_NUMBER; i++) {
				if ((next_dep_entry_index = Get_DevicesExpandersPMs_EntryIndex(pACB, i)) != ID_NOT_MAPPED) {
					return next_dep_entry_index;
				}
			}
		}
	}
	return ID_NOT_MAPPED;
}
/*
*******************************************************************************
**
*******************************************************************************
*/
PCore_Context arcsas_core_get_context_from_pool(PCore_Extension pCore)
{
	PCore_Context pContext;
	//unsigned long flags;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	/*remove the lock in case of segmentation fault in customer kernel 2.6.19-18gtc */
	arcsas_spin_lock(&pCore->context_list_lock);
	if (osm_list_empty(&pCore->Context_List)) {
		arcsas_spin_unlock(&pCore->context_list_lock);
		return NULL;
	}
	pContext = ((PCore_Context)osm_list_first_entry(&pCore->Context_List, struct _CORE_CONTEXT, Queue_Pointer));
	osm_list_del_init(&pContext->Queue_Pointer);
	/*Reset some variables.*/
	pContext->Context_Type = CORE_CONTEXT_TYPE_NONE;
	arcsas_spin_unlock(&pCore->context_list_lock);
	return pContext;
}
/*
*******************************************************************************
**
*******************************************************************************
*/
static void arcsas_free_core_context_to_pool(PCore_Extension pCore, PCore_Context *pContext)
{
	//unsigned long flags;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	arcsas_spin_lock(&pCore->context_list_lock);
	osm_list_add_tail(&(*pContext)->Queue_Pointer, &pCore->Context_List);
	*pContext = NULL;
	arcsas_spin_unlock(&pCore->context_list_lock);
	return;
}
/*
*******************************************************************************
**
*******************************************************************************
*/
struct CommandControlBlock *arcsas_core_get_coreccb_from_list(PCore_Extension pCore)
{
	struct CommandControlBlock *pCCB;
	struct AdapterControlBlock *pACB;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	pACB = pCore->pACB;
	arcsas_spin_lock(&pCore->ccb_list_lock);
	if (osm_list_empty(&pCore->Core_CCB_List)) {
		if (printk_ratelimit())
			printk("ccb pool empty\n ");
		arcsas_spin_unlock(&pCore->ccb_list_lock);
		return NULL;
	}
	pCCB = (struct CommandControlBlock *)osm_list_first_entry(&pCore->Core_CCB_List, struct CommandControlBlock, Queue_Pointer);
	#if ARCSAS_DBG_CONTEXT
		osm_printk("%s pCCB = 0x%p\n", __func__, pCCB);
	#endif
	osm_list_del_init(&pCCB->Queue_Pointer);
	arcsas_zero_ccb(pCCB);
	pCCB->CoreContext = arcsas_core_get_context_from_pool(pCore);
	arcsas_spin_unlock(&pCore->ccb_list_lock);
	return pCCB;
}
/*
*******************************************************************************
**
*******************************************************************************
*/
void arcsas_core_free_coreccb_to_pool(PCore_Extension pCore, struct CommandControlBlock *pCCB)
{
	//unsigned long flags;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	if (!IS_A_SMP_REQ(pCCB)) {
		arcsas_spin_lock(&pCCB->pDevice->sent_ccb_list_lock);
		osm_list_del_init(&pCCB->Queue_Pointer);
		arcsas_spin_unlock(&pCCB->pDevice->sent_ccb_list_lock);
	} else {
		arcsas_spin_lock(&pCore->ccb_waiting_lst_lock);
		osm_list_del_init(&pCCB->Queue_Pointer);
		arcsas_spin_unlock(&pCore->ccb_waiting_lst_lock);
	}
	arcsas_spin_lock(&pCore->ccb_list_lock);
	osm_list_add_tail(&pCCB->Queue_Pointer, &pCore->Core_CCB_List);
	if (pCCB->CoreContext) {
		arcsas_free_core_context_to_pool(pCore, &pCCB->CoreContext);
	}
	arcsas_spin_unlock(&pCore->ccb_list_lock);
	#if ARCSAS_DBG_CONTEXT
		osm_printk("%s pCCB = 0x%p\n", __func__, pCCB);
	#endif
	return;
}
/*
*******************************************************************************
**
*******************************************************************************
*/
static void arcsas_teardown_dma_pool(struct AdapterControlBlock *pACB)
{
	unsigned char core_index;
	struct _Core_Extension *pCore;

	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	switch (pACB->system_device_id) {
		case DEVICE_ID_1300: {
			pCore = pACB->pCore[0];
			if (pCore->sg_tbl_pool != NULL) {
				pci_pool_free(pCore->sg_tbl_pool, pCore->sg_tbl_pool_vaddr, pCore->sg_tbl_pool_dma);
				pci_pool_destroy(pCore->sg_tbl_pool);
			}
			if (pCore->sas_scrtch_pool != NULL) {
				pci_pool_free(pCore->sas_scrtch_pool, pCore->sas_scrtch_pool_vaddr, pCore->sas_scrtch_pool_dma);
				pci_pool_destroy(pCore->sas_scrtch_pool);
			}
			if (pCore->cmpl_queue_pool != NULL) {
				pci_pool_free(pCore->cmpl_queue_pool, pCore->cmpl_queue_pool_vaddr, pCore->cmpl_queue_pool_dma);
				pci_pool_destroy(pCore->cmpl_queue_pool);
			}
			if (pCore->post_queue_pool != NULL) {
				pci_pool_free(pCore->post_queue_pool, pCore->post_queue_pool_vaddr, pCore->post_queue_pool_dma);
				pci_pool_destroy(pCore->post_queue_pool);
			}
			if (pCore->cmnd_tbl_pool != NULL) {
				pci_pool_free(pCore->cmnd_tbl_pool, pCore->cmnd_tbl_pool_vaddr, pCore->cmnd_tbl_pool_dma);
				pci_pool_destroy(pCore->cmnd_tbl_pool);
			}
			if (pCore->rcv_fis_pool != NULL) {
				pci_pool_free(pCore->rcv_fis_pool, pCore->rcv_fis_pool_vaddr, pCore->rcv_fis_pool_dma);
				pci_pool_destroy(pCore->rcv_fis_pool);
			}
			if (pCore->cmnd_list_pool != NULL) {
				pci_pool_free(pCore->cmnd_list_pool, pCore->cmnd_list_pool_vaddr, pCore->cmnd_list_pool_dma);
				pci_pool_destroy(pCore->cmnd_list_pool);
			}
			break;
		}
 		case DEVICE_ID_1320: {
			for (core_index = 0; core_index < 2; core_index++) {
				pCore = pACB->pCore[core_index];
				if (pCore->sg_tbl_pool != NULL) {
					pci_pool_free(pCore->sg_tbl_pool, pCore->sg_tbl_pool_vaddr, pCore->sg_tbl_pool_dma);
					pci_pool_destroy(pCore->sg_tbl_pool);
				}
				if (pCore->sas_scrtch_pool != NULL) {
					pci_pool_free(pCore->sas_scrtch_pool, pCore->sas_scrtch_pool_vaddr, pCore->sas_scrtch_pool_dma);
					pci_pool_destroy(pCore->sas_scrtch_pool);
				}
				if (pCore->cmpl_queue_pool != NULL) {
					pci_pool_free(pCore->cmpl_queue_pool, pCore->cmpl_queue_pool_vaddr, pCore->cmpl_queue_pool_dma);
					pci_pool_destroy(pCore->cmpl_queue_pool);
				}
				if (pCore->post_queue_pool != NULL) {
					pci_pool_free(pCore->post_queue_pool, pCore->post_queue_pool_vaddr, pCore->post_queue_pool_dma);
					pci_pool_destroy(pCore->post_queue_pool);
				}
				if (pCore->cmnd_tbl_pool != NULL) {
					pci_pool_free(pCore->cmnd_tbl_pool, pCore->cmnd_tbl_pool_vaddr, pCore->cmnd_tbl_pool_dma);
					pci_pool_destroy(pCore->cmnd_tbl_pool);
				}
				if (pCore->rcv_fis_pool != NULL) {
					pci_pool_free(pCore->rcv_fis_pool, pCore->rcv_fis_pool_vaddr, pCore->rcv_fis_pool_dma);
					pci_pool_destroy(pCore->rcv_fis_pool);
				}
				if (pCore->cmnd_list_pool != NULL) {
					pci_pool_free(pCore->cmnd_list_pool, pCore->cmnd_list_pool_vaddr, pCore->cmnd_list_pool_dma);
					pci_pool_destroy(pCore->cmnd_list_pool);
				}
			}	
			break;
		}		
	}
}
bool arcsas_tag_is_empty(PTag_Stack pTagStack)
{
	if (pTagStack->Top==0) {
		return true;
	}
	return false;
}
/*
*******************************************************************************
** Tag: index of DevicesExpandersPMs_EntryIndex[]
** PTag_Stack: Port_Tag_Pool, Expander_Tag_Pool, PM_Tag_Pool, CmdSlot_Tag_Pool, Device_Tag_Pool
**
** if TagStackType!=FIFO_TAG, use FILO,
** if TagStackType==FIFO_TAG, use FIFO, PtrOut is the next tag to get
** and Top is the number of available tags in the stack
** when use FIFO, get tag from PtrOut and free tag to (PtrOut+Top)%Size
*******************************************************************************
*/
unsigned short arcsas_get_pool_tag(PTag_Stack pTagStack)
{
	unsigned short nTag;
	//unsigned long flags;

	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	arcsas_spin_lock(&pTagStack->lock);
	if (pTagStack->TagStackType == FIFO_TAG) {
		nTag = pTagStack->Stack[pTagStack->PtrOut++];
		if (pTagStack->PtrOut >= pTagStack->Size) {
			pTagStack->PtrOut = 0;
		}
		pTagStack->Top--;
	} else {
		nTag = pTagStack->Stack[--pTagStack->Top];
	}
	arcsas_spin_unlock(&pTagStack->lock);
	return nTag;
}
/*
*******************************************************************************
**
*******************************************************************************
*/
void arcsas_init_tag(PTag_Stack pTagStack, unsigned short size)
{
	unsigned short i;

	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	arcsas_spin_lock_init(&pTagStack->lock);
	pTagStack->Top = size;
	pTagStack->TagStackType = FILO_TAG;
	pTagStack->PtrOut = 0;
	for (i = 0; i < size; i++) {
		pTagStack->Stack[i] = size - 1 - i;
	}
	return;
}
/*
*******************************************************************************
**
*******************************************************************************
*/
void arcsas_init_cmdslot(PTag_Stack pTagStack, unsigned short size)
{
	unsigned short i;

	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	arcsas_spin_lock_init(&pTagStack->lock);
	pTagStack->Top = size;
	pTagStack->TagStackType = FIFO_TAG;
	pTagStack->PtrOut = 0;
	for (i = 0; i < size; i++) {
		pTagStack->Stack[i] = i;
	}
	return;
}
/*
*******************************************************************************
**
*******************************************************************************
*/
void arcsas_release_pool_tag(PTag_Stack pTagStack, unsigned short tag)
{
	//unsigned long flags;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	arcsas_spin_lock(&pTagStack->lock);
	if (pTagStack->TagStackType == FIFO_TAG) {
		pTagStack->Stack[(pTagStack->PtrOut + pTagStack->Top) % pTagStack->Size] = tag;
		pTagStack->Top++;
	} else {
		pTagStack->Stack[pTagStack->Top++] = tag;
	}
	arcsas_spin_unlock(&pTagStack->lock);
	return;
}
/*
*******************************************************************************
**
*******************************************************************************
*/
static void *arcsas_core_get_cached_memory(struct AdapterControlBlock *pACB, u32 size)
{
	void *mem_ptr = NULL;

	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	mem_ptr = osm_vmaloc(size + 4096);
	return mem_ptr;
}
/*
*******************************************************************************
**
*******************************************************************************
*/
static bool arcsas_core_get_DMA_memory(struct AdapterControlBlock *pACB, unsigned short PRDBufferCount, unsigned short slotCount, unsigned short deliverQueueSize, unsigned short completeQueueSize, unsigned short SASScratchCount)
{
	unsigned short i;
	Assigned_Uncached_Memory dmaResource;
	PCore_Extension pCore = NULL;
	PSAS_Scratch_Buffer pSASSB;
	PPRD_Buffer pPRD_Buffer;
	ARCSAS_U64 memDMA;
	u32 offset, size;
	void *memVir;

	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	switch (pACB->system_device_id) {
		case DEVICE_ID_1300: {
			void *vaddr;
			u32 max_hba_cmd_sg_prd_buffer_size, max_rx_fis_pool_size;
			osm_dma_addr_t dma_addr;
			struct dma_pool *dma_pool_temp;
			max_hba_cmd_sg_prd_buffer_size = MAX_HBA_CMD_3G_PRD_BUFFER_SIZE;
			max_rx_fis_pool_size = MAX_RX_FIS_POOL_SIZE_1300;
			pCore = pACB->pCore[0];
			/*
			********************************************************************************
			** Assign uncached memory for command list(64 byte align)
			********************************************************************************
			*/
			size = sizeof(struct _ARCSAS_Command_Header) * slotCount;/* Command List*/
			dma_pool_temp = pci_pool_create("arcsas command list pool", pACB->pcidev, size, 64, 0);
			if (dma_pool_temp == NULL) {
				osm_printk("Fail to create command list DMA pool.\n");
				return false;
			}
			#if ARCSAS_DBG_CONTEXT
				osm_printk("%s:got command list pool\n", __func__);
			#endif
			pCore->cmnd_list_pool = dma_pool_temp;
			vaddr = pci_pool_alloc(pCore->cmnd_list_pool, GFP_ATOMIC, &dma_addr);
			dmaResource.Virtual_Address = vaddr;
			dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
			dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr)); 
			dmaResource.Byte_Size = size;
			pCore->cmnd_list_pool_dma = dma_addr;
			pCore->cmnd_list_pool_vaddr = vaddr;
			memVir = dmaResource.Virtual_Address;
			memDMA = dmaResource.Physical_Address;
			offset = (u32)(osm_roundup(memDMA.value, 64) - memDMA.value);
			#if ARCSAS_DBG_CONTEXT
				osm_printk("%s:command list pool offset = 0x%x\n", __func__, offset);
			#endif
			memDMA = arcsas_u64_add_u32(memDMA, offset);
			memVir = (unsigned char *)memVir + offset;
			osm_memset(memVir, 0, dmaResource.Byte_Size);
			pCore->CmdSlot_DMA = memDMA;
			pCore->CmdSlot = memVir;
			/*
			********************************************************************************
			** Assign uncached memory for received FIS(256 byte align)
			********************************************************************************
			*/
			size = max_rx_fis_pool_size;/* Received FIS */
			dma_pool_temp = pci_pool_create("arcsas received FIS pool", pACB->pcidev, size, 256, 0);
			if (dma_pool_temp == NULL) {
				osm_printk("Fail to create received FIS DMA pool.\n");
				return false;
			}
			#if ARCSAS_DBG_CONTEXT
				osm_printk("%s:got received FIS pool\n", __func__);
			#endif
			pCore->rcv_fis_pool = dma_pool_temp;
			vaddr = pci_pool_alloc(pCore->rcv_fis_pool, GFP_ATOMIC, &dma_addr); 
			dmaResource.Virtual_Address = vaddr;
			dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
			dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
			dmaResource.Byte_Size = size;
			pCore->rcv_fis_pool_dma = dma_addr;
			pCore->rcv_fis_pool_vaddr = vaddr;
			memVir = dmaResource.Virtual_Address;
			memDMA = dmaResource.Physical_Address;
			offset = (u32)(osm_roundup(memDMA.value, 256) - memDMA.value);
			#if ARCSAS_DBG_CONTEXT
				osm_printk("%s:received FIS pool offset = 0x%x\n", __func__, offset);
			#endif
			memDMA = arcsas_u64_add_u32(memDMA, offset);
			memVir = (unsigned char *)memVir + offset;
			osm_memset(memVir, 0, dmaResource.Byte_Size);
			pCore->RX_FIS_DMA = memDMA;
			pCore->RX_FIS = memVir;
			/*
			*********************************************************************************
			** Assign the 32 command tables.(128 byte align)
			*********************************************************************************
			*/
			size = sizeof(struct _ARCSAS_Command_Table) * slotCount;	/* Command Table */
			dma_pool_temp = pci_pool_create("arcsas command table pool", pACB->pcidev, size, 128, 0);
			if (dma_pool_temp == NULL) {
				osm_printk("Fail to create command table DMA pool.\n");
				return false;
			}
			#if ARCSAS_DBG_CONTEXT
				osm_printk("%s:got command table pool\n", __func__);
			#endif
			pCore->cmnd_tbl_pool = dma_pool_temp;
			vaddr = pci_pool_alloc(pCore->cmnd_tbl_pool, GFP_ATOMIC, &dma_addr);
			dmaResource.Virtual_Address = vaddr;
			dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
			dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
			pCore->cmnd_tbl_pool_dma = dma_addr;
			pCore->cmnd_tbl_pool_vaddr = vaddr;
			dmaResource.Byte_Size = size;
			memVir = dmaResource.Virtual_Address;
			memDMA = dmaResource.Physical_Address;
			offset = (u32)(osm_roundup(memDMA.value, 128) - memDMA.value);
			#if ARCSAS_DBG_CONTEXT
				osm_printk("%s:command table offset = 0x%x\n", __func__, offset);
			#endif
			memDMA = arcsas_u64_add_u32(memDMA, offset);
			memVir = (unsigned char *)memVir + offset;
			pCore->Cmd_Table = memVir;
			pCore->Cmd_Table_DMA = memDMA;
			/*
			*********************************************************************************
			** Assign delivery queue.
			*********************************************************************************
			*/
			size = sizeof(struct _POST_Q_ENTRY) * deliverQueueSize;
			dma_pool_temp = pci_pool_create("arcsas post queue pool", pACB->pcidev, size, 8, 0);
			if (dma_pool_temp == NULL) {
				osm_printk("Fail to create post queue DMA pool.\n");
				return false;
			}
			#if ARCSAS_DBG_CONTEXT
				osm_printk("%s:got post queue pool\n", __func__);
			#endif
			pCore->post_queue_pool = dma_pool_temp;
			vaddr = pci_pool_alloc(pCore->post_queue_pool, GFP_ATOMIC, &dma_addr);
			dmaResource.Virtual_Address = vaddr;
			dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
			dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
			dmaResource.Byte_Size = size;
			pCore->post_queue_pool_dma = dma_addr;
			pCore->post_queue_pool_vaddr = vaddr;
			memVir = dmaResource.Virtual_Address;
			memDMA = dmaResource.Physical_Address;
			offset = (u32)(osm_roundup(memDMA.value, 8) - memDMA.value);
			#if ARCSAS_DBG_CONTEXT
				osm_printk("%s: post queue offset = 0x%x\n", __func__, offset);
			#endif
			memDMA = arcsas_u64_add_u32(memDMA, offset);
			memVir = (unsigned char *)memVir + offset;
			pCore->PostQ = memVir;/* odin delivery window of scsi,sata,sas... command */
			pCore->PostQ_DMA = memDMA;
			/*
			*********************************************************************************
			** Assign completion queue.
			*********************************************************************************
			*/
			size = sizeof(struct _DONE_Q_ENTRY) * completeQueueSize;
			dma_pool_temp = pci_pool_create("arcsas completion queue pool", pACB->pcidev, size, 8, 0);
			if (dma_pool_temp == NULL) {
				osm_printk("Fail to create completion queue DMA pool.\n");
				return false;
			}
			#if ARCSAS_DBG_CONTEXT
				osm_printk("%s:got completion queue pool\n", __func__);
			#endif
			pCore->cmpl_queue_pool = dma_pool_temp;
			vaddr = pci_pool_alloc(pCore->cmpl_queue_pool, GFP_ATOMIC, &dma_addr);
			dmaResource.Virtual_Address = vaddr;
			dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
			dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
			dmaResource.Byte_Size = size;
			pCore->cmpl_queue_pool_dma = dma_addr;
			pCore->cmpl_queue_pool_vaddr = vaddr;
			memVir = dmaResource.Virtual_Address;
			memDMA = dmaResource.Physical_Address;
			offset = (u32)(osm_roundup(memDMA.value, 8) - memDMA.value);
			#if ARCSAS_DBG_CONTEXT
				osm_printk("%s:completion queue offset = 0x%x\n", __func__, offset);
			#endif
			memDMA = arcsas_u64_add_u32(memDMA, offset);
			memVir = (unsigned char *)memVir + offset;
			pCore->DoneQ_DMA = memDMA;
			pCore->DoneQ = memVir; 
			/*
			*******************************************************
			** Assign the scratch buffer(8 byte align)
			** Number of Taret/Device can not be determined until 
			*******************************************************
			*/
			size = SAS_SCRATCH_BUFFER_SIZE * SASScratchCount;
			dma_pool_temp = pci_pool_create("arcsas sas scratch buffer pool", pACB->pcidev, size, 8, 0);
			if (dma_pool_temp == NULL) {
				osm_printk("Fail to create sas scratch buffer DMA pool.\n");
				return false;
			}
			#if ARCSAS_DBG_CONTEXT
				osm_printk("%s:got sas scratch buffer pool\n", __func__);
			#endif
			pCore->sas_scrtch_pool = dma_pool_temp;
			vaddr = pci_pool_alloc(pCore->sas_scrtch_pool, GFP_ATOMIC, &dma_addr);
			dmaResource.Virtual_Address = vaddr;
			dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
			dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
			dmaResource.Byte_Size = size;
			pCore->sas_scrtch_pool_dma = dma_addr;
			pCore->sas_scrtch_pool_vaddr = vaddr;
			memVir = dmaResource.Virtual_Address;
			memDMA = dmaResource.Physical_Address;
			offset = (u32)(osm_roundup(memDMA.value, 8) - memDMA.value);
			#if ARCSAS_DBG_CONTEXT
				osm_printk("%s:sas scratch buffer offset = 0x%x\n", __func__, offset);
			#endif
			memDMA = arcsas_u64_add_u32(memDMA, offset);
			memVir = (unsigned char *)memVir + offset;
			osm_memset(memVir, 0, dmaResource.Byte_Size);
			for (i = 0; i < SASScratchCount; i++) {
				pSASSB = (PSAS_Scratch_Buffer)osm_list_first_entry(&pCore->SAS_Scratch_List, struct _SAS_Scratch_Buffer, Queue_Pointer);
				/*$$$$$$$$$$$$$$$$*/
				pSASSB->Scratch_Vir = memVir;
				pSASSB->Scratch_DMA = memDMA;
				/*$$$$$$$$$$$$$$$$*/
				osm_list_move_tail(&pSASSB->Queue_Pointer, &pCore->SAS_Scratch_List);
				memVir = (unsigned char *)memVir + SAS_SCRATCH_BUFFER_SIZE;
				memDMA = arcsas_u64_add_u32(memDMA, SAS_SCRATCH_BUFFER_SIZE);
			}
			/*
			*********************************************************************************
			** Assign the PRD buffer 
			*********************************************************************************
			*/
			size = max_hba_cmd_sg_prd_buffer_size * PRDBufferCount;
			dma_pool_temp = pci_pool_create("arcsas sg table pool", pACB->pcidev, size, 8, 0);
			if (dma_pool_temp == NULL) {
				osm_printk("Fail to create sg table DMA pool.\n");
				return false;
			}
			#if ARCSAS_DBG_CONTEXT
				osm_printk("%s:got sg table pool\n", __func__);
			#endif
			pCore->sg_tbl_pool = dma_pool_temp;
			vaddr = pci_pool_alloc(pCore->sg_tbl_pool, GFP_ATOMIC, &dma_addr);
			dmaResource.Virtual_Address = vaddr;
			dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
			dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
			dmaResource.Byte_Size = size;
			pCore->sg_tbl_pool_dma = dma_addr;
			pCore->sg_tbl_pool_vaddr = vaddr;
			memVir = dmaResource.Virtual_Address;
			memDMA = dmaResource.Physical_Address;
			offset = (u32)(osm_roundup(memDMA.value, 8) - memDMA.value);
			#if ARCSAS_DBG_CONTEXT
				osm_printk("%s:sg table offset = 0x%x\n", __func__, offset);
			#endif
			memDMA = arcsas_u64_add_u32(memDMA, offset);
			memVir = (unsigned char *)memVir + offset;
			osm_memset(memVir, 0, dmaResource.Byte_Size);
			for (i = 0; i < PRDBufferCount; i++) {
				pPRD_Buffer = (PPRD_Buffer)osm_list_first_entry(&pCore->PRD_Buffer_List, struct _PRD_Buffer, Queue_Pointer);
				/*$$$$$$$$$$$$$$$$*/
				pPRD_Buffer->PRD_VIRTUAL = memVir;
				pPRD_Buffer->PRD_PHYSICAL = memDMA;
				/*$$$$$$$$$$$$$$$$*/
				osm_list_move_tail(&pPRD_Buffer->Queue_Pointer, &pCore->PRD_Buffer_List);
				memDMA = arcsas_u64_add_u32(memDMA, max_hba_cmd_sg_prd_buffer_size);
				memVir = (unsigned char *)memVir + max_hba_cmd_sg_prd_buffer_size; 
			}
		}
		break;
		case DEVICE_ID_1320: {
			u32 size = 0; 
			unsigned short core_index;
			u32 max_hba_cmd_sg_prd_buffer_size, max_rx_fis_pool_size;
			osm_dma_addr_t dma_addr;
			void *vaddr;
			struct dma_pool *dma_pool_temp;
			max_hba_cmd_sg_prd_buffer_size = MAX_HBA_CMD_6G_PRD_BUFFER_SIZE;
			max_rx_fis_pool_size = MAX_RX_FIS_POOL_SIZE_1320;
			for (core_index = 0; core_index < ARCSAS_MAX_CORE_NUMBER; core_index++) {
				pCore = pACB->pCore[core_index];
				/*
				******************************************************************
				**		Deploy Command Header DMA
				****************************************************************** 
				*/
				size = sizeof(struct _ARCSAS_Command_Header) * slotCount;	/* Command Header*/
				dma_pool_temp = pci_pool_create("arcsas command list pool", pACB->pcidev, size, 64, 0);
				if (dma_pool_temp == NULL) {
					osm_printk("Fail to create command list DMA pool.\n");
					return false;
				}
				#if ARCSAS_DBG_CONTEXT
					osm_printk("%s:got command list pool\n", __func__);
				#endif
				pCore->cmnd_list_pool = dma_pool_temp;
				vaddr = pci_pool_alloc(pCore->cmnd_list_pool, GFP_ATOMIC, &dma_addr);
				dmaResource.Virtual_Address = vaddr;
				dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
				dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr)); 
				dmaResource.Byte_Size = size;
				pCore->cmnd_list_pool_dma = dma_addr;
				pCore->cmnd_list_pool_vaddr = vaddr;
				memVir = dmaResource.Virtual_Address;
				memDMA = dmaResource.Physical_Address;
				offset = (u32)(osm_roundup(memDMA.value, 64) - memDMA.value);
				memDMA = arcsas_u64_add_u32(memDMA, offset);
				memVir = (unsigned char *)memVir + offset;
				osm_memset(memVir, 0, dmaResource.Byte_Size);
				pCore->CmdSlot_DMA = memDMA;
				pCore->CmdSlot = memVir;
				/*
				******************************************************************
				**		Deploy Received FIS DMA
				****************************************************************** 
				*/
				size = max_rx_fis_pool_size;/* Received FIS */
				dma_pool_temp = pci_pool_create("arcsas received FIS pool", pACB->pcidev, size, 256, 0);
				if (dma_pool_temp == NULL) {
					osm_printk("Fail to create received FIS DMA pool.\n");
					return false;
				}
				#if ARCSAS_DBG_CONTEXT
					osm_printk("%s:got received FIS pool\n", __func__);
				#endif
				pCore->rcv_fis_pool = dma_pool_temp;
				vaddr = pci_pool_alloc(pCore->rcv_fis_pool, GFP_ATOMIC, &dma_addr); 
				dmaResource.Virtual_Address = vaddr;
				dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
				dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
				dmaResource.Byte_Size = size;
				pCore->rcv_fis_pool_dma = dma_addr;
				pCore->rcv_fis_pool_vaddr = vaddr;
				memVir = dmaResource.Virtual_Address;
				memDMA = dmaResource.Physical_Address;
				offset = (u32)(osm_roundup(memDMA.value, 256) - memDMA.value);
				memDMA = arcsas_u64_add_u32(memDMA, offset);
				memVir = (unsigned char *)memVir + offset;
				osm_memset(memVir, 0, dmaResource.Byte_Size);
				pCore->RX_FIS_DMA = memDMA;
				pCore->RX_FIS = memVir;
				/*
				******************************************************************
				**		Deploy Command Table DMA
				******************************************************************
				*/
				size = sizeof(struct _ARCSAS_Command_Table) * slotCount;	/* Command Table */
				dma_pool_temp = pci_pool_create("arcsas command table pool", pACB->pcidev, size, 128, 0);
				if (dma_pool_temp == NULL) {
					osm_printk("Fail to create command table DMA pool.\n");
					return false;
				}
				#if ARCSAS_DBG_CONTEXT
					osm_printk("%s:got command table pool\n", __func__);
				#endif
				pCore->cmnd_tbl_pool = dma_pool_temp;
				vaddr = pci_pool_alloc(pCore->cmnd_tbl_pool, GFP_ATOMIC, &dma_addr);
				dmaResource.Virtual_Address = vaddr;
				dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
				dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
				pCore->cmnd_tbl_pool_dma = dma_addr;
				pCore->cmnd_tbl_pool_vaddr = vaddr;
				dmaResource.Byte_Size = size;
				memVir = dmaResource.Virtual_Address;
				memDMA = dmaResource.Physical_Address;
				offset = (u32)(osm_roundup(memDMA.value, 128) - memDMA.value);
				memDMA = arcsas_u64_add_u32(memDMA, offset);
				memVir = (unsigned char *)memVir + offset;
				osm_memset(memVir, 0, dmaResource.Byte_Size);
				pCore->Cmd_Table_DMA = memDMA;
				pCore->Cmd_Table = memVir;
				/*
				******************************************************************
				**		Deploy Post Queue DMA
				******************************************************************
				*/
				size = sizeof(struct _POST_Q_ENTRY) * deliverQueueSize;
				dma_pool_temp = pci_pool_create("arcsas post queue pool", pACB->pcidev, size, 8, 0);
				if (dma_pool_temp == NULL) {
					osm_printk("Fail to create post queue DMA pool.\n");
					return false;
				}
				#if ARCSAS_DBG_CONTEXT
					osm_printk("%s:got post queue pool\n", __func__);
				#endif
				pCore->post_queue_pool = dma_pool_temp;
				vaddr = pci_pool_alloc(pCore->post_queue_pool, GFP_ATOMIC, &dma_addr);
				dmaResource.Virtual_Address = vaddr;
				dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
				dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
				dmaResource.Byte_Size = size;
				pCore->post_queue_pool_dma = dma_addr;
				pCore->post_queue_pool_vaddr = vaddr;
				memVir = dmaResource.Virtual_Address;
				memDMA = dmaResource.Physical_Address;
				offset = (u32)(osm_roundup(memDMA.value, 8) - memDMA.value);
				memDMA = arcsas_u64_add_u32(memDMA, offset);
				memVir = (unsigned char *)memVir + offset;
				osm_memset(memVir, 0, dmaResource.Byte_Size);
				pCore->PostQ_DMA = memDMA;
				pCore->PostQ = memVir;
				/*
				******************************************************************
				**		Deploy Completion Queue DMA
				******************************************************************
				*/
				size = sizeof(struct _DONE_Q_ENTRY) * completeQueueSize;
				dma_pool_temp = pci_pool_create("arcsas completion queue pool", pACB->pcidev, size, 8, 0);
				if (dma_pool_temp == NULL) {
					osm_printk("Fail to create completion queue DMA pool.\n");
					return false;
				}
				#if ARCSAS_DBG_CONTEXT
					osm_printk("%s:got completion queue pool\n", __func__);
				#endif
				pCore->cmpl_queue_pool = dma_pool_temp;
				vaddr = pci_pool_alloc(pCore->cmpl_queue_pool, GFP_ATOMIC, &dma_addr);
				dmaResource.Virtual_Address = vaddr;
				dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
				dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
				dmaResource.Byte_Size = size;
				pCore->cmpl_queue_pool_dma = dma_addr;
				pCore->cmpl_queue_pool_vaddr = vaddr;
				memVir = dmaResource.Virtual_Address;
				memDMA = dmaResource.Physical_Address;
				offset = (u32)(osm_roundup(memDMA.value, 8) - memDMA.value);
				memDMA = arcsas_u64_add_u32(memDMA, offset);
				memVir = (unsigned char *)memVir + offset;
				osm_memset(memVir, 0, dmaResource.Byte_Size);
				pCore->DoneQ_DMA = memDMA;
				pCore->DoneQ = memVir;
				/*
				******************************************************************
				**		Deploy SAS Scratch Buffer DMA
				******************************************************************
				*/
				size = SAS_SCRATCH_BUFFER_SIZE * SASScratchCount;
				dma_pool_temp = pci_pool_create("arcsas sas scratch buffer pool", pACB->pcidev, size, 8, 0);
				if (dma_pool_temp == NULL) {
					osm_printk("Fail to create sas scratch buffer DMA pool.\n");
					return false;
				}
				#if ARCSAS_DBG_CONTEXT
					osm_printk("%s:got sas scratch buffer pool\n", __func__);
				#endif
				pCore->sas_scrtch_pool = dma_pool_temp;
				vaddr = pci_pool_alloc(pCore->sas_scrtch_pool, GFP_ATOMIC, &dma_addr);
				dmaResource.Virtual_Address = vaddr;
				dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
				dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
				dmaResource.Byte_Size = size;
				pCore->sas_scrtch_pool_dma = dma_addr;
				pCore->sas_scrtch_pool_vaddr = vaddr;
				memVir = dmaResource.Virtual_Address;
				memDMA = dmaResource.Physical_Address;
				offset = (u32)(osm_roundup(memDMA.value, 8) - memDMA.value);
				memDMA = arcsas_u64_add_u32(memDMA, offset);
				memVir = (unsigned char *)memVir + offset;
				osm_memset(memVir, 0, dmaResource.Byte_Size);
				for (i = 0; i < SASScratchCount; i++) {
					//head = arcsas_list_get_first(&pCore->SAS_Scratch_List);
					pSASSB = (PSAS_Scratch_Buffer)osm_list_first_entry(&pCore->SAS_Scratch_List, struct _SAS_Scratch_Buffer, Queue_Pointer);
					/*$$$$$$$$$$$$$$$$*/
					pSASSB->Scratch_Vir = memVir;
					pSASSB->Scratch_DMA = memDMA;
					/*$$$$$$$$$$$$$$$$*/
					osm_list_move_tail(&pSASSB->Queue_Pointer, &pCore->SAS_Scratch_List);
					memVir = (unsigned char *)memVir + SAS_SCRATCH_BUFFER_SIZE;
					memDMA = arcsas_u64_add_u32(memDMA, SAS_SCRATCH_BUFFER_SIZE);
				}
				/*
				******************************************************************
				**		Deploy SG Table DMA and SG Table DMA Trash Bucket
				******************************************************************
				*/
				size = max_hba_cmd_sg_prd_buffer_size * PRDBufferCount;
				dma_pool_temp = pci_pool_create("arcsas sg table pool", pACB->pcidev, size, 8, 0);
				if (dma_pool_temp == NULL) {
					osm_printk("Fail to create sg table DMA pool.\n");
					return false;
				}
				#if ARCSAS_DBG_CONTEXT
					osm_printk("%s:got sg table pool\n", __func__);
				#endif
				pCore->sg_tbl_pool = dma_pool_temp;
				vaddr = pci_pool_alloc(pCore->sg_tbl_pool, GFP_ATOMIC, &dma_addr);
				dmaResource.Virtual_Address = vaddr;
				dmaResource.Physical_Address.parts.low = ARCSAS_CPU_TO_LE32(dma_addr_lo32(dma_addr));
				dmaResource.Physical_Address.parts.high = ARCSAS_CPU_TO_LE32(dma_addr_hi32(dma_addr));
				dmaResource.Byte_Size = size;
				pCore->sg_tbl_pool_dma = dma_addr;
				pCore->sg_tbl_pool_vaddr = vaddr;
				memVir = dmaResource.Virtual_Address;
				memDMA = dmaResource.Physical_Address;
				offset = (u32)(osm_roundup(memDMA.value, 8) - memDMA.value);
				memDMA = arcsas_u64_add_u32(memDMA, offset);
				memVir = (unsigned char *)memVir + offset;
				osm_memset(memVir, 0, dmaResource.Byte_Size);
				for (i = 0; i < PRDBufferCount; i++) {
					pPRD_Buffer = (PPRD_Buffer)osm_list_first_entry(&pCore->PRD_Buffer_List, struct _PRD_Buffer, Queue_Pointer);
					pPRD_Buffer->PRD_VIRTUAL = memVir;
					pPRD_Buffer->PRD_PHYSICAL = memDMA;
					osm_list_move_tail(&pPRD_Buffer->Queue_Pointer, &pCore->PRD_Buffer_List);
					memVir = (unsigned char *)memVir + max_hba_cmd_sg_prd_buffer_size;
					memDMA = arcsas_u64_add_u32(memDMA, max_hba_cmd_sg_prd_buffer_size);
				}
			}
		}
		break;
		default: {
			_ASSERT();
		}
	}
	return true;
}
/*
*******************************************************************************
*This is based on Core_ModuleInitialize() in Linux driver, Odin,
*******************************************************************************
*/
static bool arcsas_core_initialize(struct AdapterControlBlock *pACB)
{
	unsigned char devCount, expanderCount, pmCount, EventElementCount;
	unsigned short slotCount, i, deliverQueueSize, completeQueueSize, max_core_num, core_index;
	unsigned short CoreCCBCount;
	unsigned short SASScratchCount, SMPScratchCount;
	unsigned short contextCount, PRDBufferCount;
	u32 discovery_buffer_size;
	u32 sgSize = 0;
	u32 size, usedSize;
	unsigned int sgEntryCount;
	PCore_Extension pCore = NULL;
	PCore_Context pContext;
	PSAS_Scratch_Buffer pSASSB;
	PSMP_Scratch_Buffer pSMPSB;
	PDomain_Port port;
	PPRD_Buffer pPRD_Buffer;
	void *pResource, *temp, *tmpSG;
	struct CommandControlBlock *pCCB;

	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	/*
	****************************************************************************************
	**			allocate RESOURCE_CACHED_MEMORY of pCore
	****************************************************************************************
	*/
	size = arcsas_core_get_cached_memory_quota(pACB);
	pResource = arcsas_core_get_cached_memory(pACB, size);
	if (pResource == NULL) {
		osm_printk("Areca sas controller core initialize, allocate core resource error.\n");
		return false;
	}
	/*
	** Zero core driver extension and all extension of cached memory field.....
	*/
	osm_memset(pResource, 0, size);
	/*
	**************Init variable
	*/
	devCount = expanderCount = pmCount = CoreCCBCount = sgEntryCount = SASScratchCount = SMPScratchCount = 0;
	contextCount = PRDBufferCount = slotCount = deliverQueueSize = completeQueueSize = EventElementCount = 0;
	/*
	**************Decide values based on driver mode *******************************************
	*/
	arcsas_get_core_supported_counts(&EventElementCount,
					&devCount, 
					&expanderCount,
					&pmCount,
					&CoreCCBCount,
					&sgEntryCount,
					&PRDBufferCount,
					&SASScratchCount,
					&SMPScratchCount,
					&slotCount,
					&contextCount,
					&deliverQueueSize,
					&completeQueueSize);
	pACB->Expander_Count_Supported = expanderCount;/*8*/
	pACB->Device_Count_Supported = devCount;/*136*/
	pACB->PortMutiplier_Count_Supported = pmCount; /*8*/
	/*
	**************************slice allocated memory for target lun Domain_Device
	*/ 
	pACB->DevicesLun = (PDomain_Device)pResource;	
	/*
	**************************slice allocated memory for target id Domain_Device
	*/
	temp = pResource + osm_roundup(sizeof(Domain_Device) * (ARCSAS_MAX_TARGETLUN + 1), 8);
	pACB->Devices = (PDomain_Device)temp;
	/*
	**************************slice allocated memory for Domain_Expander
	*/
	temp += osm_roundup(sizeof(Domain_Device) * devCount, 8);
	pACB->Expanders = (PDomain_Expander)temp;
	/*
	**************************slice allocated memory for Domain_PM
	*/
	temp += osm_roundup(sizeof(Domain_Expander) * expanderCount, 8);
	pACB->PMs = (PDomain_PM)temp;
	/*
	***************************slice allocated memory for Running_CCB
	*/
	temp += osm_roundup(sizeof(Domain_PM) * pmCount, 8);
	pACB->EventElement = (PEVENT_ELEMENT)temp;
	/* 
	**************************slice allocate memory for EventElement 
	*/    
   	 temp += osm_roundup(sizeof(struct _EVENT_ELEMENT) * EventElementCount, 8);
	/*
	*******************Device Pool*************************************************
	*/
	pACB->Device_Tag_Pool.Stack = (unsigned short *)temp;
	pACB->Device_Tag_Pool.Size = devCount;
	temp += sizeof(struct _Tag_Stack) * devCount;
	arcsas_init_tag(&pACB->Device_Tag_Pool, pACB->Device_Tag_Pool.Size);
	/*
	*******************Expander Pool************************************************
	*/
	pACB->Expander_Tag_Pool.Stack = (unsigned short *)temp;
	pACB->Expander_Tag_Pool.Size = expanderCount;
	temp += sizeof(struct _Tag_Stack) * expanderCount;
	arcsas_init_tag(&pACB->Expander_Tag_Pool, pACB->Expander_Tag_Pool.Size);
	/*
	*******************PM Pool*****************************************************
	*/
	pACB->PM_Tag_Pool.Stack = (unsigned short *)temp;
	pACB->PM_Tag_Pool.Size = pmCount;
	temp += sizeof(struct _Tag_Stack) * pmCount;
	arcsas_init_tag(&pACB->PM_Tag_Pool, pACB->PM_Tag_Pool.Size);
	/*
	*******************Init device lun structure*****************************************
	*/
	for (i = 0; i <= ARCSAS_MAX_TARGETLUN; i++) {
		pACB->DevicesLun[i].Status = DEVICE_STATUS_NO_DEVICE;
		pACB->DevicesLun[i].State = DEVICE_STATE_IDLE;
		osm_atomic_set(&pACB->DevicesLun[i].Timer_Status, STOP_RUNNING);
		pACB->DevicesLun[i].register_set = ID_NOT_MAPPED;
		pACB->DevicesLun[i].EntryIndex = ID_NOT_MAPPED;/*this value is (target_lun,target_id) */
		pACB->DevicesLun[i].Timeout_Count = 0;
		pACB->DevicesLun[i].Retry_Count = 0;
		pACB->DevicesLun[i].saslun[1] = 0x00;
	}
	/*
	*******************Init device structure*****************************************
	*/
	for (i = 0; i <= pACB->Device_Count_Supported; i++) {/*0...136=136*/
		pACB->Devices[i].Status = DEVICE_STATUS_NO_DEVICE;
		pACB->Devices[i].State = DEVICE_STATE_IDLE;
		osm_atomic_set(&pACB->Devices[i].Timer_Status, STOP_RUNNING);
		pACB->Devices[i].EntryIndex = ID_NOT_MAPPED;
		pACB->Devices[i].Timeout_Count = 0;
		pACB->Devices[i].register_set = ID_NOT_MAPPED;
		pACB->Devices[i].Retry_Count = 0;
		pACB->Devices[i].saslun[1] = 0x00;
	}
	pACB->last_device_mapped_id = 0;/*(0..135)*/
	/*
	********************Init device map structure, here we have to use maximum number*******
	*/
	for (i = 0; i <= ARCSAS_MAX_PM_ID; i++) {/*(0..135,136..143,144..151)*/
		pACB->DevicesExpandersPMs_EntryIndex[i] = ID_NOT_MAPPED;
	}
	/*
	********************Init Expander data structure************************************
	*/
	for (i = 0; i < expanderCount; i++) {
		osm_atomic_set(&pACB->Expanders[i].Device_Number, 0);
		pACB->Expanders[i].Phy_Count = 0;
		pACB->Expanders[i].pParent = NULL;
		pACB->Expanders[i].pPort = (void *)NULL;
		OSM_INIT_LIST_HEAD(&pACB->Expanders[i].Exp_Queue_Pointer);
		OSM_INIT_LIST_HEAD(&pACB->Expanders[i].Device_List);
	}
	pACB->last_expander_mapped_id = ARCSAS_MIN_EXPANDER_ID;/*136....143*/
	/*
	*******************Init PM data structure
	*/
	pACB->last_portmutiplier_mapped_id = ARCSAS_MIN_PM_ID;/*144....151*/
	/*
	********************
	*/
	max_core_num = (pACB->acb_flag & ACB_FLAG_DUAL_CORE) ? 2 : 1;
	for (core_index = 0; core_index < max_core_num; core_index++) {
		pCore = pACB->pCore[core_index] = (PCore_Extension)temp;
		temp = NumToPtr(PtrToNum(pCore) + osm_roundup(sizeof(Core_Extension), 8));
		/*
		**************Set up controller information
		*/
		pCore->pACB = pACB;
		pCore->Timer_Status = STOP_RUNNING;
		OSM_INIT_LIST_HEAD(&pCore->Wait_RegisterSet_CCB_List);
		OSM_INIT_LIST_HEAD(&pCore->CCB_Waiting_List);
		OSM_INIT_LIST_HEAD(&pCore->CCB_Complete_List);
		OSM_INIT_LIST_HEAD(&pCore->Context_List);
		OSM_INIT_LIST_HEAD(&pCore->SAS_Scratch_List);
		OSM_INIT_LIST_HEAD(&pCore->SMP_Scratch_List);
		OSM_INIT_LIST_HEAD(&pCore->PRD_Buffer_List);
		OSM_INIT_LIST_HEAD(&pCore->Core_CCB_List);
		arcsas_spin_lock_init(&pCore->ccb_cmpl_list_lock);
		arcsas_spin_lock_init(&pCore->ccb_waiting_lst_lock);
		arcsas_spin_lock_init(&pCore->ccb_list_lock);
		arcsas_spin_lock_init(&pCore->context_list_lock);
		arcsas_spin_lock_init(&pCore->prd_list_lock);
		arcsas_spin_lock_init(&pCore->sassb_list_lock);
		arcsas_spin_lock_init(&pCore->smpsb_list_lock);
		arcsas_spin_lock_init(&pCore->doneq_lock);
		arcsas_spin_lock_init(&pCore->eh_lock);
		tasklet_init(&pCore->isr_tasklet, arcsas_interrupt_service_bh, (unsigned long)pCore);
		pCore->Slot_Count_Supported = slotCount;/*512*/
		/*
		**************For deliver queue, queue size is the entry size. *********************
		*/
		pCore->Max_PostQ_Index = deliverQueueSize;
		/* 
		********************************************************************************************
		** For completion queue,
		** Minus 1 because the first entry in completion queue is the pointer not real entry
		********************************************************************************************
		*/
		pCore->Max_DoneQ_Index = completeQueueSize - 1;//522
		switch (pACB->system_device_id) {
			case DEVICE_ID_1300: {
				pCore->core_flag |= ARCSAS_CORE_FLAG_HBA_CORE_ID0;
				pCore->Phy_Num = 4;
				pCore->MaxRegisterSet = 16;
				pCore->last_port_index = 0;
				pCore->MaxCmdSlotWidth = 9;
				#if ARCSAS_DBG_CONTEXT
					osm_printk("DEVICE_ID [%x] is found. Rev %x\n", pACB->system_device_id, pACB->Revision_Id);
				#endif
				break;
			}
			case DEVICE_ID_1320: {
				pCore->core_flag |= (core_index ? ARCSAS_CORE_FLAG_HBA_CORE_ID1 : ARCSAS_CORE_FLAG_HBA_CORE_ID0);
				pCore->Phy_Num = 4;
				pCore->MaxRegisterSet = 64;
				pCore->last_port_index = 0;
				pCore->MaxCmdSlotWidth = 9;
				#if ARCSAS_DBG_CONTEXT
					osm_printk("DEVICE_ID [%x] is found. Rev %x\n", pACB->system_device_id, pACB->Revision_Id);
				#endif
				break;
			}
			default: {
				_ASSERT();
			}
		}
		/*
		*************************************************************
		*************************************************************
		*/
		pCore->Running_CCB = (PCCB *)temp;
		/*
		***************************slice allocate memory for CCB pool
		***************************load ccb sg entry pointer
		***************************link ccb pool of Core_CCB_List
		*/
		temp += osm_roundup(sizeof(struct CommandControlBlock *) * slotCount, 8);
		sgSize = sizeof(struct _ARCSAS_SGENTRY) * sgEntryCount;
		tmpSG = temp;
		temp = temp + sgSize * CoreCCBCount;
		for (i = 0; i < CoreCCBCount; i++) {/*CoreCCBCount=128*/
			pCCB = (struct CommandControlBlock *)temp;
			pCCB->SG_Table.pSGEntry = (PARCSAS_SGENTRY)tmpSG;
			osm_list_add_tail(&pCCB->Queue_Pointer, &pCore->Core_CCB_List);	/*link all ccb pointer*/
			tmpSG += sgSize;
			temp += sizeof(struct CommandControlBlock);			/*struct _CCB is 64byte aligned.*/
		}
		/*
		************struct _CORE_CONTEXT pool
		************link pCore pContext pool of Context_List
		*/
		temp = NumToPtr(osm_roundup(PtrToNum(temp), 8));
		for (i = 0; i < contextCount; i++) {
			pContext = (PCore_Context)temp;
			pContext->Context_Type = CORE_CONTEXT_TYPE_NONE;
			OSM_INIT_LIST_HEAD(&pContext->Queue_Pointer);
			osm_list_add_tail(&pContext->Queue_Pointer, &pCore->Context_List);
			temp += sizeof(struct _CORE_CONTEXT);
		}
		/*
		************struct _SAS_Scratch_Buffer pool
		************link pCore pSASSB pool of SAS_Scratch_List
		*/
		temp = NumToPtr(osm_roundup(PtrToNum(temp), 8));
		for (i = 0; i < SASScratchCount; i++) {
			pSASSB = (PSAS_Scratch_Buffer)temp;
			OSM_INIT_LIST_HEAD(&pSASSB->Queue_Pointer);
			osm_list_add_tail(&pSASSB->Queue_Pointer, &pCore->SAS_Scratch_List);
			temp += sizeof(struct _SAS_Scratch_Buffer);
		}
		/*
		************struct _SMP_Scratch_Buffer pool
		************link pCore pSMPSB pool of SMP_Scratch_List
		*/
		temp = NumToPtr(osm_roundup(PtrToNum(temp), 8));
		for (i = 0; i < SMPScratchCount; i++) {
			pSMPSB = (PSMP_Scratch_Buffer)temp;
			OSM_INIT_LIST_HEAD(&pSMPSB->Queue_Pointer);
			osm_list_add_tail(&pSMPSB->Queue_Pointer, &pCore->SMP_Scratch_List);
			temp += sizeof(struct _SMP_Scratch_Buffer);
		}
		/*
		************struct _PRD_Buffer pool
		************link pCore pPRD_Buffer pool of PRD_Buffer_List
		*/
		temp = NumToPtr(osm_roundup(PtrToNum(temp), 8));
		for (i = 0; i < PRDBufferCount; i++) {/*sgBufferCount = 512*/
			pPRD_Buffer = (PPRD_Buffer)temp;
			OSM_INIT_LIST_HEAD(&pPRD_Buffer->Queue_Pointer);
			osm_list_add_tail(&pPRD_Buffer->Queue_Pointer, &pCore->PRD_Buffer_List);
			temp += sizeof(struct _PRD_Buffer);
		}
		/*
		**********************************************************************************
		** Initialize pools: Tag pools
		** Special case here: In hibernation we must handle only one request
		** at a time, but we also need 2 slots in order for completion queue
		** entry pointer to work properly. Therefore, slotCount will be
		** allocated to 2 during hibernation but we only give one tag in the
		** pool; same for XOR slots
		***********************************************************************************
		*/
		pCore->CmdSlot_Tag_Pool.Stack = (unsigned short *)temp;
		pCore->CmdSlot_Tag_Pool.Size = slotCount;
		temp += sizeof(struct _Tag_Stack) * slotCount;
		arcsas_init_cmdslot(&pCore->CmdSlot_Tag_Pool, pCore->CmdSlot_Tag_Pool.Size);
		/*
		*******************Port Pool*****************************************************
		*/
		pCore->Port_Tag_Pool.Stack = (unsigned short *)temp;
		pCore->Port_Tag_Pool.Size = ARCSAS_MAX_CORE_PHYSICAL_PORT_NUMBER;
		temp += sizeof(struct _Tag_Stack) * ARCSAS_MAX_CORE_PHYSICAL_PORT_NUMBER;
		arcsas_init_tag(&pCore->Port_Tag_Pool, ARCSAS_MAX_CORE_PHYSICAL_PORT_NUMBER);
		/*
		**************************************************************************************
		*****	Port_MAP and last_port_index will be read from the register
		*****	Init port data structure
		**************************************************************************************
		*/
		for (i = 0; i < ARCSAS_MAX_CORE_PHYSICAL_PORT_NUMBER; i++) {
			port = &pCore->Ports[i];
			port->Port_Index = (unsigned char) i;
			port->Port_State = PORT_STATE_IDLE;
			osm_atomic_set(&port->DiscoveryInstance, 0);
			port->pCore = pCore;
			osm_atomic_set(&port->Device_Number, 0);
			osm_atomic_set(&port->Expander_Number, 0);
			OSM_INIT_LIST_HEAD(&port->Device_List);
			OSM_INIT_LIST_HEAD(&port->Expander_List);
			/* PORT TYPE will be determined after arcsas_init_chip. */
			port->Type = PORT_TYPE_SAS; /*default*/
		}
		pCore->last_port_index = 0;
		/*
		********************Init device map structure
		*/
		for (i = 0; i < ARCSAS_MAX_CORE_PHYSICAL_PORT_NUMBER; i++) {
			pCore->Ports_EntryIndex[i] = ID_NOT_MAPPED;
		}
		for (i = 0; i < 16; i++) {
			pCore->Running_Slot[i] = 0;
			pCore->Resetting_Slot[i] = 0;
		}
		pCore->last_PostQ_index = 0xffff;
		pCore->last_DoneQ_index = 0xffff;
	}
	/*
	**************************************************************************************
	******	Rest of memory will be used for discovery
	**************************************************************************************
	*/
	usedSize = (u32)(PtrToNum(temp) - PtrToNum(pResource));
	discovery_buffer_size = (pACB->acb_flag & ACB_FLAG_DUAL_CORE) ? ((size - usedSize) >> 1) : (size - usedSize);
	pACB->pCore[0]->pDiscoverBuffer = temp;
	arcsas_discovery_set_resource(pACB->pCore[0]->pDiscoverBuffer, size - usedSize, expanderCount);
	if (pACB->acb_flag & ACB_FLAG_DUAL_CORE) {
		pACB->pCore[1]->pDiscoverBuffer = NumToPtr(PtrToNum(temp) + discovery_buffer_size);
		arcsas_discovery_set_resource(pACB->pCore[1]->pDiscoverBuffer, discovery_buffer_size, expanderCount);
	}
	if (!arcsas_core_get_DMA_memory(pACB, PRDBufferCount, slotCount, deliverQueueSize, completeQueueSize, SASScratchCount)) {
		arcsas_teardown_dma_pool(pACB);
		return false;
	}
	return true;
}
/*
*******************************************************************************
** This function stops the core driver from carrying out any more requests
** It disables all interrupts in the hardware and aborts all requests that
** are either running or waiting in the core driver
*******************************************************************************
*/
static void arcsas_hba_shutdown(PACB pACB)
{
	unsigned int max_core_num, core_index;
	PCore_Extension pCore;
	unsigned char i;

	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	if (pACB->acb_flag & ACB_FLAG_CORES_STATE_STARTED) {
		max_core_num = (pACB->acb_flag & ACB_FLAG_DUAL_CORE) ? 2 : 1;
		for (core_index = 0; core_index < max_core_num; core_index++) {
			pCore = pACB->pCore[core_index];
			for (i = 0; i < ARCSAS_MAX_CORE_PHYSICAL_PORT_NUMBER; i++) {
				unsigned char index;
				index = pCore->Ports_EntryIndex[i];
				if (index != ID_NOT_MAPPED) {
					arcsas_port_abort_ccbs(&pCore->Ports[index], CCB_STATUS_INVALID_REQUEST, NULL);
				}
			}
		}
		pACB->acb_flag &= ~ACB_FLAG_CORES_STATE_STARTED;
	}
	return;
}
/*
*******************************************************************************
**
*******************************************************************************
*/
void arcsas_pre_init_pm(PDomain_PM pPM)
{
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	osm_memset((unsigned char *)pPM, 0, sizeof(Domain_PM));
	pPM->Timer_Status = STOP_RUNNING;
	arcsas_init_timer(&pPM->timer_request);
	pPM->register_set = ID_NOT_MAPPED;
	OSM_INIT_LIST_HEAD(&pPM->Sent_CCB_List);
	arcsas_spin_lock_init(&pPM->sent_ccb_list_lock);
	arcsas_spin_lock_init(&pPM->timer_lock);
	return;
}
/*
*******************************************************************************
**
*******************************************************************************
*/
static void arcsas_pre_init_device(PDomain_Device pDevice)
{
	struct work_struct add_work_buffer, remove_work_buffer;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __FUNCTION__);
	#endif
	if ((pDevice->Status & DEVICE_STATUS_SHOWED_ONCE) == 0) {
		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 20)
			OSM_INIT_WORK(&pDevice->arcsas_add_scsi_device, arcsas_add_scsi_device_wqfn);
			OSM_INIT_WORK(&pDevice->arcsas_remove_scsi_device, arcsas_remove_scsi_device_wqfn);
		#else
			OSM_INIT_WORK(&pDevice->arcsas_add_scsi_device, arcsas_add_scsi_device_wqfn, pDevice);
			OSM_INIT_WORK(&pDevice->arcsas_remove_scsi_device, arcsas_remove_scsi_device_wqfn, pDevice);
		#endif
		pDevice->Status = DEVICE_STATUS_NO_DEVICE;
	} else {
		add_work_buffer = pDevice->arcsas_add_scsi_device;
		remove_work_buffer = pDevice->arcsas_remove_scsi_device;
		osm_memset((unsigned char *)pDevice, 0, sizeof(Domain_Device));
		pDevice->arcsas_add_scsi_device = add_work_buffer;
		pDevice->arcsas_remove_scsi_device = remove_work_buffer;
		pDevice->Status = DEVICE_STATUS_NO_DEVICE | DEVICE_STATUS_SHOWED_ONCE;
	}
	osm_atomic_set(&pDevice->Timer_Status, STOP_RUNNING);
	arcsas_init_timer(&pDevice->timer_request);
	pDevice->Queue_Depth = 64;//Originally this is 255
	pDevice->State = DEVICE_STATE_IDLE;
	pDevice->register_set = ID_NOT_MAPPED;
	pDevice->Timeout_Count = 0;
	pDevice->Retry_Count = 0;
	OSM_INIT_LIST_HEAD(&pDevice->Port_Queue_Pointer);
	OSM_INIT_LIST_HEAD(&pDevice->Exp_Queue_Pointer);
	OSM_INIT_LIST_HEAD(&pDevice->Sent_CCB_List);
	arcsas_spin_lock_init(&pDevice->sent_ccb_list_lock);
	arcsas_spin_lock_init(&pDevice->ncq_lock);
	arcsas_spin_lock_init(&pDevice->timer_lock);
	return;
}
/*
*******************************************************************************
**
*******************************************************************************
*/
PDomain_Device arcsas_get_device_lun_from_pool(PCore_Extension pCore, unsigned short EntryIndex)
{
	unsigned char i = 0;
	struct AdapterControlBlock *pACB;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	pACB = pCore->pACB;
	for (i = 0; i < ARCSAS_MAX_TARGETLUN; i++) {
		if (pACB->DevicesLun[i].saslun[1] == 0x00 && (pACB->DevicesLun[i].EntryIndex == ID_NOT_MAPPED)) {
			/* saslun=0 imply that this Device_Lun is free */
			break;
		}
	}
	if (i > ARCSAS_MAX_TARGETLUN) {
		#if ARCSAS_DBG_CONTEXT
			osm_printk("Ran out of Lun device IDs!\n");
		#endif
		return NULL;
	}
	arcsas_pre_init_device(&pACB->DevicesLun[i]);
	pACB->DevicesLun[i].pCore = pCore;
	pACB->DevicesLun[i].EntryIndex = EntryIndex;/*this value is (target_lun,target_id) */
	pACB->DevicesLun[i].State = DEVICE_STATE_IDLE;
	/* Set some default value. */
	pACB->DevicesLun[i].MDMA_Mode = 0xFF;
	pACB->DevicesLun[i].PIO_Mode = 0xFF;
	pACB->DevicesLun[i].UDMA_Mode = 0xFF;
	pACB->DevicesLun[i].Current_MDMA = 0xFF;
	pACB->DevicesLun[i].Current_PIO = 0xFF;
	pACB->DevicesLun[i].Current_UDMA = 0xFF;
	pACB->DevicesLun[i].NegotiatedLinkRate = 0;
	osm_atomic_inc(&pACB->all_device_count);
	return &pACB->DevicesLun[i];
}

/*
*******************************************************************************
** for DT_DIRECT_ACCESS_BLOCK,DT_SEQ_ACCESS,DT_PRINTER
**     DT_PROCESSOR,DT_WRITE_ONCE,DT_CD_DVD,DT_OPTICAL_MEMORY
**	   DT_MEDIA_CHANGER,DT_STORAGE_ARRAY_CTRL,DT_ENCLOSURE
**     DT_EXPANDER,DT_PM,DT_SES_DEVICE
*******************************************************************************
*/
PDomain_Device arcsas_get_device_from_pool(PCore_Extension pCore)
{
	unsigned short index;
	unsigned char i = 0;
	struct AdapterControlBlock *pACB;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	pACB = pCore->pACB;
	if (arcsas_tag_is_empty(&pACB->Device_Tag_Pool)) {
		osm_printk("Device pool empty!\n");
		return NULL;
	}
	index = arcsas_get_pool_tag(&pACB->Device_Tag_Pool);/*Tag: dep_entry_index of DevicesExpandersPMs_EntryIndex[]*/
	/*
	**********************************************************************
	** for consistant binding, we will try to find the first available
	** ID this ID is already taken - we need to find another
	** should only happen after a lot of hot plugs, so that the
	** IDs looped around
	**********************************************************************
	*/
	for (i = 0; i <= ARCSAS_MAX_DEVICE_ID; i++) {/*0...135*/
		if (pACB->DevicesExpandersPMs_EntryIndex[i] == ID_NOT_MAPPED) {
			/* update the last ID into last_device_mapped_id */
			if (pACB->last_device_mapped_id < i) {
				pACB->last_device_mapped_id = i;/*0...135*/
			}
			break;
		}
	}
	if (i > ARCSAS_MAX_DEVICE_ID) {
		osm_printk("Ran out of device IDs!\n");
		return NULL;
	}
	pACB->DevicesExpandersPMs_EntryIndex[i] = (unsigned char)index;/*0....135*/
	arcsas_pre_init_device(&pACB->Devices[index]);
	pACB->Devices[index].EntryIndex = i;/*0...135*/
	pACB->Devices[index].pCore = pCore;
	pACB->Devices[index].State = DEVICE_STATE_IDLE;
	/* Set some default value. */
	pACB->Devices[index].MDMA_Mode = 0xFF;
	pACB->Devices[index].PIO_Mode = 0xFF;
	pACB->Devices[index].UDMA_Mode = 0xFF;
	pACB->Devices[index].Current_MDMA = 0xFF;
	pACB->Devices[index].Current_PIO = 0xFF;
	pACB->Devices[index].Current_UDMA = 0xFF;
	pACB->Devices[index].NegotiatedLinkRate = 0;
	osm_atomic_inc(&pACB->all_device_count);
	return &pACB->Devices[index];
}
/*
*******************************************************************************
**This is identical with generate_sg_table() in Odin.
*******************************************************************************
*/
static bool arcsas_generate_os_ccb_sg_table(struct AdapterControlBlock *pACB, struct scsi_cmnd *pscsi_cmnd, PARCSAS_SG_Table pSGTable)
{
	unsigned int i;
	unsigned int length;
	struct scatterlist *sg;
	int sg_count;
	__le32 address_lo, address_hi;

	#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_OS_CCB)
		osm_printk("%s:\n", __func__);
	#endif
	if (arcsas_rq_bflen(pscsi_cmnd) > ARCSAS_MAX_CMDXFER_LENGTH) {
		osm_printk("arcsas: request buffer length exceeds the maximum alowed value.\n");
		return false;
	}
	if (arcsas_use_sg(pscsi_cmnd)) {
		#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT)
			osm_printk("%s: use sg\n", __func__);
		#endif
		sg_count = arcsas_map_sg(pscsi_cmnd);
		#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT)
			osm_printk("%s: sg_count = %d\n", __func__, sg_count);
		#endif
		if ((unlikely(sg_count >= ARCSAS_MAX_OS_SG_ENTRIES)) || (unlikely(sg_count < 0))) {
			osm_printk("arcsas: sg dma mapping failed.\n");
			return false;
		}
		#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT)
			osm_printk("%s: use sg, sg_count = %d\n", __func__, sg_count);
		#endif
		sg = (struct scatterlist *) arcsas_rq_bf(pscsi_cmnd);
		scsi_for_each_sg(pscsi_cmnd, sg, sg_count, i) {
			address_lo = cpu_to_le32(dma_addr_lo32(sg_dma_address(sg)));
			address_hi = cpu_to_le32(dma_addr_hi32(sg_dma_address(sg)));
			length = sg_dma_len(sg);
			arcsas_sg_table_append(pSGTable, address_lo, address_hi, length);
		}
	} else if (arcsas_rq_bflen(pscsi_cmnd)) {
		dma_addr_t dma_addr;
		#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT)
			osm_printk("%s: no sg\n", __func__);
		#endif
		dma_addr = pci_map_single(pACB->pcidev, arcsas_rq_bf(pscsi_cmnd), arcsas_rq_bflen(pscsi_cmnd), pscsi_cmnd->sc_data_direction);
		/* We hide it here for later unmap.*/
		pscsi_cmnd->SCp.dma_handle = dma_addr;
		address_lo = cpu_to_le32(dma_addr_lo32(dma_addr));
		address_hi = cpu_to_le32(dma_addr_hi32(dma_addr));
		arcsas_sg_table_append(pSGTable, address_lo, address_hi, arcsas_rq_bflen(pscsi_cmnd));
	}
	return true;
}
/*
*******************************************************************************
** This should be the _only_ os request exit point
** This is identical with hba_req_callback() in Odin.
*******************************************************************************
*/
void arcsas_done_os_ccb_callback(void *this, struct CommandControlBlock *pCCB)
{
	/* Return this request to OS. */
	PDomain_Device pDevice;
	PACB pACB = (PACB)this;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	pDevice = pCCB->pDevice;
	arcsas_free_ncq_tag(pDevice, pCCB);
	arcsas_os_ccb_done(pACB, pCCB);
	return;
}
/*
*******************************************************************************
** This is identical to scsi_cmd_to_req_conv() in Odin.
*******************************************************************************
*/
static unsigned long arcsas_build_os_ccb(struct AdapterControlBlock *pACB, struct scsi_cmnd *pscsi_cmnd, PDomain_Device *ppDevice)
{
	u64 timeout;
	PCore_Extension pCore;
	unsigned short entry_index;
	struct CommandControlBlock *pOSCCB = NULL;
	/*
	*************************************************
	**Set three flags:	CMD_FLAG_NON_DATA
	**		CMD_FLAG_DATA_IN
	**		CMD_FLAG_DMA
	**Currently data in/out all go thru DMA
	*************************************************
	*/
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	if (pscsi_cmnd->device->lun) {
		unsigned char i;
		*ppDevice = NULL;
		entry_index = MAKE_DEVICE_ENTRY_INDEX(pscsi_cmnd->device->id, pscsi_cmnd->device->lun);
		for (i = 0; i <= ARCSAS_MAX_TARGETLUN; i++) {
			if (pACB->DevicesLun[i].EntryIndex == entry_index) {/*this value is (target_lun,target_id)*/
				*ppDevice = &pACB->DevicesLun[i];
				break;
			}
		}
	} else {
		entry_index = Get_DevicesExpandersPMs_EntryIndex(pACB, pscsi_cmnd->device->id);
		if (entry_index == ID_NOT_MAPPED) {
			*ppDevice = NULL;/*hot unplug, device is disconnected*/
		} else {
			*ppDevice = &pACB->Devices[entry_index];
		}
	}
	if (*ppDevice == NULL) {
		return 0;
	}
	pCore = (*ppDevice)->pCore;
	if (((*ppDevice)->Error_Handling_State == EH_PORT_RESET) || ((*ppDevice)->pPort->Setting & PORT_SETTING_DURING_RECOVERY)) {
		if (printk_ratelimit())
			osm_printk("EH_PORT_RESET or PORT_SETTING_DURING_RECOVERY, return SCSI_MLQUEUE_DEVICE_BUSY\n ");
		return SCSI_MLQUEUE_DEVICE_BUSY;
	}
	//if ((osm_atomic_read(&(*ppDevice)->pPort->DiscoveryInstance) != 0) || ((*ppDevice)->State != DEVICE_STATE_INIT_DONE)) {
	if ((*ppDevice)->State != DEVICE_STATE_INIT_DONE) {
		unsigned long rtn;
			rtn = SCSI_MLQUEUE_DEVICE_BUSY;
			if (printk_ratelimit())
				osm_printk("pDevice->State != DEVICE_STATE_INIT_DONE, return SCSI_MLQUEUE_DEVICE_BUSY\n ");
		return rtn;
	}
	pOSCCB = arcsas_core_get_coreccb_from_list(pCore);
	if (pOSCCB == NULL) {
		return 0;
	}
	arcsas_sg_table_init(&pOSCCB->SG_Table, ARCSAS_MAX_OS_SG_ENTRIES);
	if (arcsas_generate_os_ccb_sg_table(pACB, pscsi_cmnd, &pOSCCB->SG_Table)) { /*fill pCCB->SG_Table*/
		arcsas_set_lba_and_sector_count(pOSCCB);/*fill pCCB->LBA and pCCB->Sector_Count*/
		#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 27)
			timeout = pscsi_cmnd->request->timeout / HZ;//timeout value in struct request is in jiffy
			pscsi_cmnd->request->deadline = jiffies + (timeout * ((MAX_TIMEOUT_ALLOWED + 0) * 2 + 2)) * HZ;
			if (pscsi_cmnd->device->request_queue->timeout.function) {
				mod_timer(&pscsi_cmnd->device->request_queue->timeout, pscsi_cmnd->request->deadline);			
			}
		#else
			timeout = pscsi_cmnd->timeout_per_command / HZ;//timeout value in struct request is in jiffy
			if (pscsi_cmnd->eh_timeout.function) {
				mod_timer(&pscsi_cmnd->eh_timeout, jiffies + (timeout * ((MAX_TIMEOUT_ALLOWED + 0) * 2 + 2)) * HZ );
			}
		#endif
		switch (pscsi_cmnd->sc_data_direction) {
			case DMA_NONE:
				break;
			case DMA_FROM_DEVICE:
				pOSCCB->Cmd_Flag |= CMD_FLAG_DATA_IN;
				break;
			case DMA_TO_DEVICE:
				pOSCCB->Cmd_Flag |= CMD_FLAG_DMA;
				break;
			case DMA_BIDIRECTIONAL:
				osm_printk("arcsas : unexpected DMA_BIDIRECTIONAL.\n");
				break;
			default:
				break;
		}
		osm_memcpy(pOSCCB->Cdb, pscsi_cmnd->cmnd, MAX_CDB_SIZE);
		pOSCCB->EntryIndex = MAKE_DEVICE_ENTRY_INDEX(pscsi_cmnd->device->id, pscsi_cmnd->device->lun);
		pOSCCB->pDevice = *ppDevice;
		pOSCCB->PRD_Buffer = NULL;
		pOSCCB->Data_Buffer = arcsas_rq_bf(pscsi_cmnd);
		pOSCCB->Data_Transfer_Length = arcsas_rq_bflen(pscsi_cmnd);
		pOSCCB->Sense_Info_Buffer = pscsi_cmnd->sense_buffer;
		pOSCCB->Sense_Info_Buffer_Length = SCSI_SENSE_BUFFERSIZE;
		pOSCCB->Time_Out = timeout; // timeout in second
		pOSCCB->Org_Req = pscsi_cmnd;
		pOSCCB->Tag = pscsi_cmnd->tag;
		pOSCCB->CCB_Status = CCB_STATUS_PENDING;
		pOSCCB->Completion = (void(*)(void *, PCCB))arcsas_done_os_ccb_callback;
		pOSCCB->Cmd_Initiator = (void *)pACB;
	} else {
		arcsas_core_free_coreccb_to_pool(pCore, pOSCCB);
		//Although it fails to generate os ccb sg table here, it will return the ccb address to tell arcsas_queuecommand that the ccb pool is not empty.
	}
	return (unsigned long)pOSCCB;
}
/*
*******************************************************************************
** This is identical with mv_complete_request() in Odin.
*******************************************************************************
*/
void arcsas_os_ccb_done(struct AdapterControlBlock *pACB, struct CommandControlBlock *pCCB)
{
	struct scsi_cmnd *pscsi_cmnd;
	PARCSAS_Sense_Data  senseBuffer;
	PDomain_Device pDevice;
	PCore_Extension pCore;

	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	pscsi_cmnd = (struct scsi_cmnd *)pCCB->Org_Req;
	senseBuffer = (PARCSAS_Sense_Data) pscsi_cmnd->sense_buffer;
	pDevice = pCCB->pDevice;
	pCore = pDevice->pCore;
	switch (pCCB->CCB_Status) {
		case CCB_STATUS_SUCCESS:
			pscsi_cmnd->result = DID_OK << 16;
			#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT)
				osm_printk("%s: pscsi_cmnd = 0x%p, pCCB = 0x%p CCB_STATUS_SUCCESS, DID_OK\n", __func__, pscsi_cmnd, pCCB);
			#endif
			break;
		case CCB_STATUS_MEDIA_ERROR:
			pscsi_cmnd->result = DID_PARITY << 16;
			#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT)
				osm_printk("%s: pscsi_cmnd = 0x%p, pCCB = 0x%p CCB_STATUS_MEDIA_ERROR, DID_ABORT\n", __func__, pscsi_cmnd, pCCB);
			#endif
			break;
		case CCB_STATUS_BUSY:
			pscsi_cmnd->result = DID_BUS_BUSY << 16;
			#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT)
				osm_printk("%s: pscsi_cmnd = 0x%p, pCCB = 0x%p CCB_STATUS_BUSY, DID_BUS_BUSY\n", __func__, pscsi_cmnd, pCCB);
			#endif
			break;
		case CCB_STATUS_NO_DEVICE:
			pscsi_cmnd->result = DID_NO_CONNECT << 16;
			#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT || ARCSAS_DBG_PORT_ABORT_CCB)
				osm_printk("%s: pscsi_cmnd = 0x%p, pCCB = 0x%p CCB_STATUS_NO_DEVICE, DID_NO_CONNECT\n", __func__, pscsi_cmnd, pCCB);
				osm_printk("%s: pCore->Completing_Slot[0x%x] = 0x%x,  pCore->Running_Slot[0x%x] = 0x%x\n", __func__, pCCB->SlotNo >> 5, pCore->Completing_Slot[pCCB->SlotNo >> 5], pCCB->SlotNo >> 5, pCore->Running_Slot[pCCB->SlotNo >> 5]);
				osm_printk("%s: pCore->Running_CCB[0x%x] = 0x%p\n", __func__, pCCB->SlotNo, pCore->Running_CCB[pCCB->SlotNo]);
			#endif
			break;
		case CCB_STATUS_HAS_SENSE://The ccb has sense data inside
			#if ARCSAS_DBG_HAS_SENSE
				osm_printk("%s: pCCB = 0x%p, CCB_STATUS_HAS_SENSE\n", __func__, pCCB);
				osm_printk("%s: scsi_cmnd(0x%p)[0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x], id:lun = %x:%x\n", __func__,
					pscsi_cmnd,
					pscsi_cmnd->cmnd[0],
					pscsi_cmnd->cmnd[1],
					pscsi_cmnd->cmnd[2],
					pscsi_cmnd->cmnd[3],
					pscsi_cmnd->cmnd[4],
					pscsi_cmnd->cmnd[5],
					pscsi_cmnd->cmnd[6],
					pscsi_cmnd->cmnd[7],
					pscsi_cmnd->cmnd[8],
					pscsi_cmnd->cmnd[9],
					pscsi_cmnd->device->id,
					pscsi_cmnd->device->lun
				);
			#endif
			pscsi_cmnd->result = DRIVER_SENSE << 24 | DID_OK << 16 | CHECK_CONDITION << 1;
			senseBuffer->ErrorCode = SCSI_SENSE_CURRENT_ERRORS;
			/* Sense buffer data is valid already. */
			senseBuffer->Valid = 1;
			switch (senseBuffer->SenseKey) {
				case SCSI_SK_RECOVERED_ERROR:
				case SCSI_SK_MEDIUM_ERROR:
					if (senseBuffer->SenseKey == SCSI_SK_RECOVERED_ERROR) {
						osm_printk("%s: SenseKey = SCSI_SK_RECOVERED_ERROR\n", __func__);
						osm_printk("%s: ASC = 0x%x, ASCQ = 0x%x, SKS[0] = 0x%x, SKS[1] = 0x%x, SKS[2] = 0x%x\n", __func__, senseBuffer->AdditionalSenseCode, senseBuffer->AdditionalSenseCodeQualifier, senseBuffer->SenseKeySpecific[0], senseBuffer->SenseKeySpecific[1], senseBuffer->SenseKeySpecific[2]);
					} else {
						osm_printk("%s: SenseKey = SCSI_SK_MEDIUM_ERROR\n", __func__);
						osm_printk("%s: ASC = 0x%x, ASCQ = 0x%x, SKS[0] = 0x%x, SKS[1] = 0x%x, SKS[2] = 0x%x\n", __func__, senseBuffer->AdditionalSenseCode, senseBuffer->AdditionalSenseCodeQualifier, senseBuffer->SenseKeySpecific[0], senseBuffer->SenseKeySpecific[1], senseBuffer->SenseKeySpecific[2]);
					}
					if (pDevice->EntryIndex != ID_NOT_MAPPED) {
						pDevice->CmdErrorCount++;
						if (pDevice->CmdErrorCount > ARCSAS_MAX_DEVICE_RECOVERED_MEDIUM_ERROR_COUNT) {
							arcsas_remove_device(pDevice);
							pscsi_cmnd->result = DID_NO_CONNECT << 16;
						}
					} else {
						pscsi_cmnd->result = DID_NO_CONNECT << 16;
					}
					break;
				case SCSI_SK_HARDWARE_ERROR:
					osm_printk("%s: SCSI_SK_HARDWARE_ERROR\n", __func__);
					osm_printk("%s: ASC = 0x%x, ASCQ = 0x%x, SKS[0] = 0x%x, SKS[1] = 0x%x, SKS[2] = 0x%x\n", __func__, senseBuffer->AdditionalSenseCode, senseBuffer->AdditionalSenseCodeQualifier, senseBuffer->SenseKeySpecific[0], senseBuffer->SenseKeySpecific[1], senseBuffer->SenseKeySpecific[2]);
					if (pDevice->EntryIndex != ID_NOT_MAPPED) {
						arcsas_remove_device(pDevice);
					}
					pscsi_cmnd->result = DID_NO_CONNECT << 16;
					break;
				case SCSI_SK_ILLEGAL_REQUEST:
					osm_printk("%s: SenseKey = SCSI_SK_ILLEGAL_REQUEST\n", __func__);
					osm_printk("%s: ASC = 0x%x, ASCQ = 0x%x, SKS[0] = 0x%x, SKS[1] = 0x%x, SKS[2] = 0x%x\n", __func__, senseBuffer->AdditionalSenseCode, senseBuffer->AdditionalSenseCodeQualifier, senseBuffer->SenseKeySpecific[0], senseBuffer->SenseKeySpecific[1], senseBuffer->SenseKeySpecific[2]);
					if ((pDevice->EntryIndex != ID_NOT_MAPPED) && (senseBuffer->AdditionalSenseCode == SCSI_ASC_LOGICAL_UNIT_NOT_SUPPORTED))/* scsi id lun */{
						if (DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex))/* scsi id lun */ {
							pDevice->Status = DEVICE_STATUS_NO_DEVICE;
							pDevice->State = DEVICE_STATE_IDLE;
							if (IS_ENCLOSURE(pDevice)) {
								arcsas_free_enclosure_entry_index(pACB, pDevice->EntryIndex);/* scsi id lun */
							}
							arcsas_port_abort_ccbs(pDevice->pPort, CCB_STATUS_NO_DEVICE, pDevice);
							if (pDevice->pExpander != NULL) {
								osm_list_del_init(&pDevice->Exp_Queue_Pointer);
								osm_atomic_dec(&pDevice->pExpander->Device_Number);
							}
							osm_list_del_init(&pDevice->Port_Queue_Pointer);
							osm_atomic_dec(&pDevice->pPort->Device_Number);
							arcsas_free_device_lun_to_pool(pCore, pDevice);
						} else {
							if (arcsas_device_babies_lun_check(pDevice)) {  
								arcsas_port_abort_ccbs(pDevice->pPort, CCB_STATUS_NO_DEVICE, pDevice);
								pDevice->Dev_Type = DT_LOGICAL_UNIT_NOT_PRESENT_DEVICE;
								pDevice->State = DEVICE_STATE_INIT_DONE;
							} else {
								arcsas_remove_device(pDevice); /*need handle if this device had LUN device*/
							}
						}
					}
					pscsi_cmnd->result = DID_NO_CONNECT << 16;
					break;
				case SCSI_SK_ABORTED_COMMAND:
					osm_printk("%s:SenseKey = SCSI_SK_ABORTED_COMMAND, ASC = 0x%x, ASCQ = 0x%x\n", __func__, senseBuffer->AdditionalSenseCode, senseBuffer->AdditionalSenseCodeQualifier);
					pscsi_cmnd->result = DID_ABORT << 16;
					break;
				case SCSI_SK_NOT_READY:
					osm_printk("%s:SenseKey = SCSI_SK_NOT_READY, ASC = 0x%x, ASCQ = 0x%x\n", __func__, senseBuffer->AdditionalSenseCode, senseBuffer->AdditionalSenseCodeQualifier);
					pscsi_cmnd->result = DID_NO_CONNECT << 16;
					break;
				default:
					osm_printk("%s: SenseKey = 0x%x, ASC = 0x%x, ASCQ = 0x%x\n", __func__, senseBuffer->SenseKey, senseBuffer->AdditionalSenseCode, senseBuffer->AdditionalSenseCodeQualifier);
					pscsi_cmnd->result = DID_REQUEUE << 16;
					break;
			}
			break;
		case CCB_STATUS_ERROR_WITH_SENSE://This ccb is from API only
			senseBuffer->ErrorCode = SCSI_SENSE_CURRENT_ERRORS;
			/* Sense buffer data is valid already. */
			osm_printk("%s: pscsi_cmnd = 0x%p, pCCB = 0x%p CCB_STATUS_ERROR_WITH_SENSE\n", __func__, pscsi_cmnd, pCCB);
			break;
		case CCB_STATUS_INVALID_REQUEST:
			pscsi_cmnd->result = DID_ABORT << 16;
			#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT)
			osm_printk("%s: pCCB = 0x%p scsi_dev = 0x%p CCB_STATUS_INVALID_REQUEST\n", __func__, pCCB, pscsi_cmnd);
			osm_printk("%s: scsi_cmnd(0x%p)[0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x], id:lun = %x:%x\n", __func__,
				pscsi_cmnd,
				pscsi_cmnd->cmnd[0],
				pscsi_cmnd->cmnd[1],
				pscsi_cmnd->cmnd[2],
				pscsi_cmnd->cmnd[3],
				pscsi_cmnd->cmnd[4],
				pscsi_cmnd->cmnd[5],
				pscsi_cmnd->cmnd[6],
				pscsi_cmnd->cmnd[7],
				pscsi_cmnd->cmnd[8],
				pscsi_cmnd->cmnd[9],
				pscsi_cmnd->device->id,
				pscsi_cmnd->device->lun
			);
			#endif
			break;
		case CCB_STATUS_ERROR:
			pscsi_cmnd->result = DID_ERROR << 16;
			#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT)
				osm_printk("%s: pscsi_cmnd = 0x%p, pCCB = 0x%p CCB_STATUS_ERROR, DID_ERROR\n", __func__, pscsi_cmnd, pCCB);
				osm_printk("%s: pCore->Completing_Slot[0x%x] = 0x%x,  pCore->Running_Slot[0x%x] = 0x%x\n", __func__, pCCB->SlotNo >> 5, pCore->Completing_Slot[pCCB->SlotNo >> 5], pCCB->SlotNo >> 5, pCore->Running_Slot[pCCB->SlotNo >> 5]);
				osm_printk("%s: pCore->Running_CCB[0x%x] = 0x%p\n", __func__, pCCB->SlotNo, pCore->Running_CCB[pCCB->SlotNo]);
			#endif
			break;
		case CCB_STATUS_DID_ABORT:
			pscsi_cmnd->result = DID_ABORT << 16;
			#if (ARCSAS_DBG_CONTEXT || ARCSAS_DBG_OS_CCB_CONTEXT || ARCSAS_DBG_PORT_ABORT_CCB)
				osm_printk("%s:pscsi_cmnd = 0x%p, ID:LUN = %d: %d pCCB = 0x%p CCB_STATUS_DID_ABORT\n", __func__, pscsi_cmnd, pscsi_cmnd->device->id, pscsi_cmnd->device->lun, pCCB);
			#endif
			break;	
		default:
			pscsi_cmnd->result = DID_ABORT << 16;
	}
	arcsas_core_free_coreccb_to_pool(pCore, pCCB);
	if (arcsas_use_sg(pscsi_cmnd)) {
		arcsas_unmap_sg(pscsi_cmnd);
	} else if (arcsas_rq_bflen(pscsi_cmnd)) {
		pci_unmap_single(pACB->pcidev, (dma_addr_t)pscsi_cmnd->SCp.dma_handle, arcsas_rq_bflen(pscsi_cmnd), pscsi_cmnd->sc_data_direction);
	}
	pscsi_cmnd->scsi_done(pscsi_cmnd);
	return;
}
/*
*********************************************************************************
*********************************************************************************
*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 37)
	static int arcsas_queuecommand_lck(struct scsi_cmnd *pscsi_cmnd, void (*done)(struct scsi_cmnd *))
	{

		struct Scsi_Host *host;
		struct AdapterControlBlock *pACB;
		unsigned long pccb;
		struct pci_dev *pcidev;
		PDomain_Device pDevice = NULL;

		#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_OS_CCB)
			osm_printk("%s: scsi_cmnd(0x%p)[0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x], id:lun = %x:%x\n", __func__,
				pscsi_cmnd,
				pscsi_cmnd->cmnd[0],
				pscsi_cmnd->cmnd[1],
				pscsi_cmnd->cmnd[2],
				pscsi_cmnd->cmnd[3],
				pscsi_cmnd->cmnd[4],
				pscsi_cmnd->cmnd[5],
				pscsi_cmnd->cmnd[6],
				pscsi_cmnd->cmnd[7],
				pscsi_cmnd->cmnd[8],
				pscsi_cmnd->cmnd[9],
				pscsi_cmnd->device->id,
				pscsi_cmnd->device->lun
			);
			mdelay(20);
		#endif
		host = pscsi_cmnd->device->host;
		pACB = (struct AdapterControlBlock *)host->hostdata;
		pscsi_cmnd->scsi_done = done;
		pscsi_cmnd->result = 0;
		pcidev = pACB->pcidev;
		if (pACB->acb_flag & ACB_FLAG_SYNC_DEVICE_ID) {
			return SCSI_MLQUEUE_HOST_BUSY;
		}
		if (unlikely(osm_atomic_read(&pACB->outstandingccbcount) > host->can_queue)) {
			if (printk_ratelimit())
				osm_printk("queuecommand: SCSI_MLQUEUE_HOST_BUSY\n ");
			return SCSI_MLQUEUE_HOST_BUSY;
		}
		if (unlikely(pscsi_cmnd->device->channel)) {
			pscsi_cmnd->result = DID_BAD_TARGET << 16;
			pscsi_cmnd->scsi_done(pscsi_cmnd);
			return 0;
		}
		if ((pscsi_cmnd->cmnd[0] == SCSI_CMD_ATA_PASSTHRU_12) || (pscsi_cmnd->cmnd[0] == SCSI_CMD_ATA_PASSTHRU_16)) {
			pscsi_cmnd->result = DID_ERROR << 16 | COMMAND_COMPLETE << 8;
			pscsi_cmnd->scsi_done(pscsi_cmnd);
			return 0;
		}
		#if ARCSAS_SUPPORT_IOCTL
			if (pscsi_cmnd->device->id == ARCSAS_VIRTUAL_DEVICE_ID) {
				arcsas_handle_virtual_ses_device(pACB, pscsi_cmnd);
				return 0;
			}
		#endif
		pccb = arcsas_build_os_ccb(pACB, pscsi_cmnd, &pDevice);
		if (pccb) {
			if ((pccb == SCSI_MLQUEUE_HOST_BUSY) || (pccb == SCSI_MLQUEUE_DEVICE_BUSY)) {
				return pccb;
			} else {
				arcsas_post_ccb(((struct CommandControlBlock *)pccb)->pDevice->pCore, (struct CommandControlBlock *)pccb);
			}
			return 0;
		} else {
			if (pDevice == NULL) {
				pscsi_cmnd->result = DID_NO_CONNECT << 16;//This is important. Driver should tell scsi-layer actively
				pscsi_cmnd->scsi_done(pscsi_cmnd);
				return 0;
			} else {
				return SCSI_MLQUEUE_HOST_BUSY;
			}
		}
	}
	DEF_SCSI_QCMD(arcsas_queuecommand)
#else
	static int arcsas_queuecommand(struct scsi_cmnd *pscsi_cmnd, void (*done)(struct scsi_cmnd *))
	{

		struct Scsi_Host *host;
		struct AdapterControlBlock *pACB;
		unsigned long pccb;
		struct pci_dev *pcidev;
		PDomain_Device pDevice = NULL;

		#if (ARCSAS_DBG_FUNCTION || ARCSAS_DBG_OS_CCB)
			osm_printk("%s: scsi_cmnd(0x%p)[0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x], id:lun = %x:%x\n", __func__,
				pscsi_cmnd,
				pscsi_cmnd->cmnd[0],
				pscsi_cmnd->cmnd[1],
				pscsi_cmnd->cmnd[2],
				pscsi_cmnd->cmnd[3],
				pscsi_cmnd->cmnd[4],
				pscsi_cmnd->cmnd[5],
				pscsi_cmnd->cmnd[6],
				pscsi_cmnd->cmnd[7],
				pscsi_cmnd->cmnd[8],
				pscsi_cmnd->cmnd[9],
				pscsi_cmnd->device->id,
				pscsi_cmnd->device->lun
			);
			mdelay(20);
		#endif
		host = pscsi_cmnd->device->host;
		pACB = (struct AdapterControlBlock *)host->hostdata;
		pscsi_cmnd->scsi_done = done;
		pscsi_cmnd->result = 0;
		pcidev = pACB->pcidev;

		if (unlikely(osm_atomic_read(&pACB->outstandingccbcount) > host->can_queue)) {
			if (printk_ratelimit())
				osm_printk("queuecommand: SCSI_MLQUEUE_HOST_BUSY\n ");
			return SCSI_MLQUEUE_HOST_BUSY;
		}
		if (unlikely(pscsi_cmnd->device->channel)) {
			pscsi_cmnd->result = DID_BAD_TARGET << 16;
			pscsi_cmnd->scsi_done(pscsi_cmnd);
			return 0;
		}
		if ((pscsi_cmnd->cmnd[0] == SCSI_CMD_ATA_PASSTHRU_12) || (pscsi_cmnd->cmnd[0] == SCSI_CMD_ATA_PASSTHRU_16)) {
			pscsi_cmnd->result = DID_ERROR << 16 | COMMAND_COMPLETE << 8;
			pscsi_cmnd->scsi_done(pscsi_cmnd);
			return 0;
		}
		#if ARCSAS_SUPPORT_IOCTL
			if (pACB->acb_flag & ACB_FLAG_SYNC_DEVICE_ID) {
				return SCSI_MLQUEUE_HOST_BUSY;
			}
			if (pscsi_cmnd->device->id == ARCSAS_VIRTUAL_DEVICE_ID) {
				arcsas_handle_virtual_ses_device(pACB, pscsi_cmnd);
				return 0;
			}
		#endif
		pccb = arcsas_build_os_ccb(pACB, pscsi_cmnd, &pDevice);
		if (pccb) {
			if ((pccb == SCSI_MLQUEUE_HOST_BUSY) || (pccb == SCSI_MLQUEUE_DEVICE_BUSY)) {
				return pccb;
			} else {
				arcsas_post_ccb(((struct CommandControlBlock *)pccb)->pDevice->pCore, (struct CommandControlBlock *)pccb);
			}
			return 0;
		} else {
			if (pDevice == NULL) {
				pscsi_cmnd->result = DID_NO_CONNECT << 16;//This is important. Driver should tell scsi-layer actively
				pscsi_cmnd->scsi_done(pscsi_cmnd);
				return 0;
			} else {
				return SCSI_MLQUEUE_HOST_BUSY;
			}
		}
	}
#endif
/*
*******************************************************************************
**
*******************************************************************************
*/
static void arcsas_pci_remove(struct pci_dev *pcidev)
{
	struct Scsi_Host *host;
	PDomain_Device pDevice;
	PDomain_PM pPM;
	struct AdapterControlBlock *pacb;
	unsigned char core_index, max_core_num;
	unsigned char i;
	unsigned short index;

	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	host = pci_get_drvdata(pcidev);
	pacb = (struct AdapterControlBlock *) host->hostdata;
	max_core_num = (pacb->acb_flag & ACB_FLAG_DUAL_CORE) ? 2 : 1;
	arcsas_hba_shutdown(pacb);
	for (i = 0; i < ARCSAS_MAX_DEVICE_ID; i++) {/*(0..135,136..143,144..151)*/
		if (pacb->DevicesExpandersPMs_EntryIndex[i] == ID_NOT_MAPPED) {
			continue;
		}
		pDevice = &pacb->Devices[Get_DevicesExpandersPMs_EntryIndex(pacb, i)];/*(0..135)*/
		arcsas_delete_device_timer(pacb, pDevice);
	}
	scsi_remove_host(host);
	scsi_host_put(host);
	free_irq(pcidev->irq, pacb);
	for (core_index = 0; core_index < max_core_num; core_index++) {
		tasklet_kill(&pacb->pCore[core_index]->isr_tasklet);
	}
	arcsas_disable_hba_isr(pacb);
	if (pacb->acb_flag & ACB_FLAG_PERSISTENT_MAP_ENABLE) {
		arcsas_check_update_flash_device_map(pacb, false);/* rom region 0x30000....0x30400, 8 bytes pattern for each device 8*128=1024 */
	}
	switch (pacb->system_device_id) {
			case DEVICE_ID_1300: {
				iounmap(pacb->pCore[0]->mem_map_base);
				break;
			}
			case DEVICE_ID_1320: {
				iounmap((pacb->pCore[0]->mem_map_base) - 0x20000);
				break;
			}
			default: {
				_ASSERT();
				break;
			}
	}
	arcsas_teardown_dma_pool(pacb);
	osm_vfree(pacb->DevicesLun);	
	for (i = ARCSAS_MIN_PM_ID; i <= ARCSAS_MAX_PM_ID; i++) {
		if (pacb->DevicesExpandersPMs_EntryIndex[i] != ID_NOT_MAPPED) {
			index = pacb->DevicesExpandersPMs_EntryIndex[i];
			pPM = &pacb->PMs[index];
			arcsas_delete_pm_timer(pacb, pPM);
		}
	}
	pci_release_regions(pcidev);
	pci_disable_device(pcidev);
	pci_set_drvdata(pcidev, NULL);
	return;
}
/*
*******************************************************************************
**
*******************************************************************************
*/
static int arcsas_os_eh_abort_handler(struct scsi_cmnd *pscsi_cmnd)
{
	PDomain_Device pDevice;
	struct CommandControlBlock *pCCB;
	PCore_Extension pCore;
	unsigned short entry_index;
	PDomain_Port pPort;
	PACB pACB;
	unsigned long t;
	unsigned short i;
	osm_printk("\n%s: scsi_cmnd(0x%p)[0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x, 0x%2x] on scsi_id = 0x%x scsi_lun = 0x%x.\n",
		__func__,
		pscsi_cmnd,
		pscsi_cmnd->cmnd[0],
		pscsi_cmnd->cmnd[1],
		pscsi_cmnd->cmnd[2],
		pscsi_cmnd->cmnd[3],
		pscsi_cmnd->cmnd[4],
		pscsi_cmnd->cmnd[5],
		pscsi_cmnd->cmnd[6],
		pscsi_cmnd->cmnd[7],
		pscsi_cmnd->cmnd[8],
		pscsi_cmnd->cmnd[9],
		pscsi_cmnd->device->id,
		pscsi_cmnd->device->lun
	);
	pACB = acb_addr;
	if (pscsi_cmnd->device->lun) {
		unsigned char i;
		pDevice = NULL;
		entry_index = MAKE_DEVICE_ENTRY_INDEX(pscsi_cmnd->device->id, pscsi_cmnd->device->lun);
		for (i = 0; i <= ARCSAS_MAX_TARGETLUN; i++) {
			if (pACB->DevicesLun[i].EntryIndex == entry_index) {/*this value is (target_lun,target_id)*/
				pDevice = &pACB->DevicesLun[i];
				break;
			}
		}
	} else {
		entry_index = Get_DevicesExpandersPMs_EntryIndex(pACB, pscsi_cmnd->device->id);
		if (entry_index == ID_NOT_MAPPED) {
			pDevice = NULL;/*hot unplug, device is disconnected*/
		} else {
			pDevice = &pACB->Devices[entry_index];
		}
	}
	if ((pDevice == NULL) || (pDevice->State == DEVICE_STATE_IDLE)) {
  		osm_printk("scsi device has been deleted\n");
		//pscsi_cmnd->result = DID_NO_CONNECT << 16;
		//if (arcsas_use_sg(pscsi_cmnd)) {
		//	arcsas_unmap_sg(pscsi_cmnd);
		//} else if (arcsas_rq_bflen(pscsi_cmnd)) {
		//	pci_unmap_single(pACB->pcidev, (osm_dma_addr_t)(unsigned long)pscsi_cmnd->SCp.dma_handle, arcsas_rq_bflen(pscsi_cmnd), pscsi_cmnd->sc_data_direction);
		//}
		//pscsi_cmnd->scsi_done(pscsi_cmnd);
		return SUCCESS;
	}
	t = jiffies;
	do {
		ssleep(5);
		if (time_after(jiffies, t + 30 * HZ))
			break;
	} while (!((pDevice->State == DEVICE_STATE_INIT_DONE) && (pDevice->Error_Handling_State == EH_NONE)));
	pCore = pDevice->pCore;
	pPort = pDevice->pPort;
	for (i = 0; i < pCore->Slot_Count_Supported; i++) {/*Slot_Count_Supported = 512*/
		pCCB = (struct CommandControlBlock *)pCore->Running_CCB[i];
		if (pCCB != NULL) {
			if (pscsi_cmnd == pCCB->Org_Req) {
				if (entry_index == pCCB->EntryIndex) {
					osm_printk("%s: found ccb(0x%p) in Running_CCB list\n", __func__, pCCB);
					pCore->Completing_Slot[pCCB->SlotNo >> 5] &= ~(1L << (pCCB->SlotNo % 32));
					pCCB->CCB_Status = CCB_STATUS_DID_ABORT;
					arcsas_complete_command(pCore, pCCB);
					pDevice->Error_Handling_State = EH_NONE;
					return SUCCESS;
				}
			}
		}
	}
	if (i >= pCore->Slot_Count_Supported) {
		osm_printk("%s: can't find ccb in Running_CCB list\n", __func__);
	}
	osm_printk("scsi_eh_abort_handler returns success for pscsicmd = 0x%p scsi_cmnd[0x%2x] on scsi_id = 0x%x, scsi_lun = 0x%x.\n",
		pscsi_cmnd,
		pscsi_cmnd->cmnd[0],
		pscsi_cmnd->device->id,
		pscsi_cmnd->device->lun
	);
	//if (arcsas_use_sg(pscsi_cmnd)) {
	//	arcsas_unmap_sg(pscsi_cmnd);
	//} else if (arcsas_rq_bflen(pscsi_cmnd)) {
	//	pci_unmap_single(pACB->pcidev, (osm_dma_addr_t)(unsigned long)pscsi_cmnd->SCp.dma_handle, arcsas_rq_bflen(pscsi_cmnd), pscsi_cmnd->sc_data_direction);
	//}
	//pscsi_cmnd->scsi_done(pscsi_cmnd);
	return SUCCESS;
}
/*
*******************************************************************************
**
*******************************************************************************
*/
#if 0
static int arcsas_os_eh_dev_reset_handler(struct scsi_cmnd *pscsi_cmnd)
{
	PDomain_Device pDevice = NULL;
	PDomain_Port pPort;
	unsigned short entry_index;
	PCore_Extension pCore;
	struct AdapterControlBlock *pACB = acb_addr;
	struct CommandControlBlock *pCCBtimeout;
	unsigned long flags;

	WARN(1, "%s OS EH: Device Reset.\n"
		"pscsicmd = 0x%p scsi_cmnd[0x%2x] on scsi_id = 0x%x scsi_lun = 0x%x.\n",
		__func__,
		pscsi_cmnd,
		pscsi_cmnd->cmnd[0],
		pscsi_cmnd->device->id,
		pscsi_cmnd->device->lun
	);
	if (pscsi_cmnd->device->lun) {
		unsigned char i;
		entry_index = MAKE_DEVICE_ENTRY_INDEX(pscsi_cmnd->device->id, pscsi_cmnd->device->lun);
		#if ARCSAS_DBG_CONTEXT
			osm_printk("%s: entry_index = 0x%x\n", __func__, entry_index);
		#endif
		for (i = 0; i <= ARCSAS_MAX_TARGETLUN; i++) {
			#if 1
				osm_printk("%s: i = %d\n", __func__, i);
			#endif
			if (pACB->DevicesLun[i].EntryIndex == entry_index) {/*this value is (target_lun,target_id)*/
				pDevice = &pACB->DevicesLun[i];
				#if 1
					osm_printk("%s: pDevice = 0x%p,  entry_index = 0x%x\n", __func__, pDevice, entry_index);
				#endif
				break;
			}
		}
	} else {
		entry_index = Get_DevicesExpandersPMs_EntryIndex(pACB, pscsi_cmnd->device->id);
		pDevice = &pACB->Devices[entry_index];
		#if 1
			osm_printk("%s:No.2\n", __func__);
			osm_printk("%s:pDevice = 0x%p\n", __func__, pDevice);
			osm_printk("%s:entry_index = 0x%d\n", __func__, entry_index);
		#endif
	}
	if (pDevice == NULL) {
		_ASSERT();
	}
	arcsas_delete_device_timer(pACB, pDevice);
	pDevice->Error_Handling_State = EH_SET_DISK_RESET;
	pCore = pDevice->pCore;
	pPort = pDevice->pPort;
	/* get the request that caused this time out */
	arcsas_spin_lock_irqsave(&pDevice->sent_ccb_list_lock, &flags);
	if (!osm_list_empty(&pDevice->Sent_CCB_List)) {/*Sent_CCB_List is not empty*/
		pCCBtimeout = (struct CommandControlBlock *)osm_list_first_entry(&pDevice->Sent_CCB_List, struct CommandControlBlock, Queue_Pointer);
		/*Add to tail, for SAS, let the next timeout to come, for SATA, it will be deleted below anyway.*/
		#if 1
			osm_printk("line %d pCCBtimeout = 0x%p\n", __LINE__, pCCBtimeout);
		#endif
		osm_list_move_tail(&pCCBtimeout->Queue_Pointer, &pDevice->Sent_CCB_List);
		/*Check if it is already handled (reset/aborted)*/
		if ((pCore->Resetting_Slot[pCCBtimeout->SlotNo/32] & (0x1L << (pCCBtimeout->SlotNo % 32))) && (!osm_list_empty(&pDevice->Sent_CCB_List))) {
			pCCBtimeout = (struct CommandControlBlock *)osm_list_first_entry(&pDevice->Sent_CCB_List, struct CommandControlBlock, Queue_Pointer);
			#if 1
				osm_printk("line %d pCCBtimeout = 0x%p\n", __LINE__, pCCBtimeout);
			#endif
			 /*To Tail, for SAS, let the next timeout to come, for SATA, it will be deleted below anyway.*/
			osm_list_move_tail(&pCCBtimeout->Queue_Pointer, &pDevice->Sent_CCB_List);
		}
	} else {
		arcsas_spin_unlock_irqrestore(&pDevice->sent_ccb_list_lock, &flags);
		return FAILED;
	}
	arcsas_spin_unlock_irqrestore(&pDevice->sent_ccb_list_lock, &flags);
	osm_printk("Command (0x%p) timeout on device 0x%p, scsi id %d, scsi lun %d.\n", pCCBtimeout, pDevice, DEVICE_ENTRY_INDEX_TO_TARGET_ID(pDevice->EntryIndex), DEVICE_ENTRY_INDEX_TO_TARGET_LUN(pDevice->EntryIndex));
	if ((Get_DevicesExpandersPMs_EntryIndex(pACB, pCCBtimeout->EntryIndex) != ID_NOT_MAPPED) && (osm_atomic_read(&pPort->DiscoveryInstance) == 0) && (pPort->Type & PORT_TYPE_SAS) && IS_SSP(pDevice)) {
		arcsas_sas_error_handling(pDevice, pCCBtimeout);
	}
	if ((pDevice != NULL) && IS_STP_OR_SATA_ATAPI(pDevice)) {
		arcsas_spin_lock_irqsave(&pDevice->sent_ccb_list_lock, &flags);
		osm_list_del_init(&pCCBtimeout->Queue_Pointer);
		arcsas_spin_unlock_irqrestore(&pDevice->sent_ccb_list_lock, &flags);
		arcsas_sata_error_handling(pCCBtimeout);
	}
	return SUCCESS;
}
/*
*******************************************************************************
**
*******************************************************************************
*/
static int arcsas_os_eh_bus_reset_handler(struct scsi_cmnd *pscsicmd)
{
	PACB pACB = (PACB)pscsicmd->device->host->hostdata;
	int max_core_num, core_index;
	struct _Domain_Device *pDevice;
	struct _Core_Extension *pCore;
	bool need_reset_bus;
	unsigned char i;
	unsigned short entryindex;
	osm_printk("scsi_eh_bus_reset_handler.pscsicmd = 0x%p scsi_cmnd[0x%2x] on scsi_id = 0x%x scsi_lun = 0x%x.\n",
		pscsicmd,
		pscsicmd->cmnd[0],
		pscsicmd->device->id,
		pscsicmd->device->lun
	);
	entryindex = Get_DevicesExpandersPMs_EntryIndex(pACB, MAKE_DEVICE_ENTRY_INDEX(pscsicmd->device->id, pscsicmd->device->lun));
	pDevice = &pACB->Devices[entryindex];
	max_core_num = (pACB->acb_flag & ACB_FLAG_DUAL_CORE) ? 2 : 1;
	for (core_index = 0; core_index < max_core_num; core_index++) {
		pCore = pACB->pCore[core_index];
		for (i = 0; i < ARCSAS_MAX_CORE_PHYSICAL_PORT_NUMBER; i++) {
			unsigned char index;
			index = pCore->Ports_EntryIndex[i];
			if (index != ID_NOT_MAPPED) {
				need_reset_bus = arcsas_port_abort_ccbs(&pCore->Ports[index], CCB_STATUS_NO_DEVICE, NULL);
			}
		}
	}
	if (need_reset_bus) {
		arcsas_hba_reset(pACB);
		arcsas_quiet_hba_isr(pACB);
		for (core_index = 0; core_index < max_core_num; core_index++) {
			pCore = pACB->pCore[core_index];
			pCore->State |= CORE_STATE_IDLE;
			arcsas_port_state_machine(pCore, NULL);
			do {
				if (arcsas_interrupt_check(pACB)) {
					osm_writel(INT_CMD_CMPL_MASK, pCore->mem_map_base + COMMON_IRQ_STATE);// clear CMD_CMPLT ASAP
					arcsas_interrupt_service_bh((unsigned long)pCore);
				}
			} while (!(pCore->State & CORE_STATE_STARTED));
		}
		while (!(pACB->acb_flag & ACB_FLAG_CORES_STATE_STARTED)) {
			for (core_index = 0; core_index < max_core_num; core_index++) {
				pCore = pACB->pCore[core_index];
				/* clear SAS interrupt */
				if (arcsas_interrupt_check(pACB)) {
					osm_writel(INT_CMD_CMPL_MASK, pCore->mem_map_base + COMMON_IRQ_STATE);// clear CMD_CMPLT ASAP
					arcsas_interrupt_service_bh((unsigned long)pCore);
				}
			}
		}
		arcsas_unquiet_hba_isr(pACB);
	}
	#if ARCSAS_DBG_CONTEXT
		osm_printk("scsi_eh_bus_reset_handler return\n");
	#endif
	return SUCCESS;
}
#endif
/*
*********************************************************************************
*********************************************************************************
*/
#if defined(RHEL_RELEASE_CODE) && LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 31)
	static int arcsas_change_queue_depth(struct scsi_device *sdev, int queue_depth, int reason)
	{
		unsigned short entry_index;
		PDomain_Device pDevice = NULL;
		int res = 0;
		if (reason != SCSI_QDEPTH_DEFAULT)
			return -EOPNOTSUPP;
		if (sdev->lun) {
			unsigned char i;
			entry_index = MAKE_DEVICE_ENTRY_INDEX(sdev->id, sdev->lun);
			for (i = 0; i <= ARCSAS_MAX_TARGETLUN; i++) {
				if (acb_addr->DevicesLun[i].EntryIndex == entry_index) {/*this value is (target_lun,target_id)*/
					pDevice = &acb_addr->DevicesLun[i];
					break;
				}
			}
		} else {
			entry_index = Get_DevicesExpandersPMs_EntryIndex(acb_addr, sdev->id);
			if (entry_index == ID_NOT_MAPPED) {
				pDevice = NULL;/*hot unplug, device is disconnected*/
			} else {
				pDevice = &acb_addr->Devices[entry_index];
			}
		}
		if (pDevice == NULL) {
			osm_printk("fail to change queue_depth of scsi id:lun(%d:%d)\n", sdev->id, sdev->lun);
			return -EOPNOTSUPP;
		}
		if (IS_SSP(pDevice)) {
			res = min(queue_depth, SAS_MAX_QD);
			if (sdev->tagged_supported)
				scsi_adjust_queue_depth(sdev, scsi_get_tag_type(sdev), res);
			else {
				scsi_adjust_queue_depth(sdev, 0, 1);
				res = 1;
			}
		} else if (IS_STP_OR_SATA_ATAPI(pDevice)) {
			res = min(queue_depth, ARCSAS_MAX_CMD_PERDEV);
			if (sdev->tagged_supported)
				scsi_adjust_queue_depth(sdev, scsi_get_tag_type(sdev), res);
			else {
				scsi_adjust_queue_depth(sdev, 0, 1);
				res = 1;
			}			
		}
		osm_printk("queue_depth of scsi id:lun(%d:%d) has changed as %d\n", res, sdev->id, sdev->lun);
		return res;
	}
#elif LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 32)
	static int arcsas_change_queue_depth(struct scsi_device *sdev, int queue_depth, int reason)
	{
		unsigned short entry_index;
		PDomain_Device pDevice = NULL;
		int res = 0;
		if (reason != SCSI_QDEPTH_DEFAULT)
			return -EOPNOTSUPP;
		if (sdev->lun) {
			unsigned char i;
			entry_index = MAKE_DEVICE_ENTRY_INDEX(sdev->id, sdev->lun);
			for (i = 0; i <= ARCSAS_MAX_TARGETLUN; i++) {
				if (acb_addr->DevicesLun[i].EntryIndex == entry_index) {/*this value is (target_lun,target_id)*/
					pDevice = &acb_addr->DevicesLun[i];
					break;
				}
			}
		} else {
			entry_index = Get_DevicesExpandersPMs_EntryIndex(acb_addr, sdev->id);
			if (entry_index == ID_NOT_MAPPED) {
				pDevice = NULL;/*hot unplug, device is disconnected*/
			} else {
				pDevice = &acb_addr->Devices[entry_index];
			}
		}
		if (pDevice == NULL) {
			osm_printk("fail to change queue_depth of scsi id:lun(%d:%d)\n", sdev->id, sdev->lun);
			return -EOPNOTSUPP;
		}
		if (IS_SSP(pDevice)) {
			res = min(queue_depth, SAS_MAX_QD);
			if (sdev->tagged_supported)
				scsi_adjust_queue_depth(sdev, scsi_get_tag_type(sdev), res);
			else {
				scsi_adjust_queue_depth(sdev, 0, 1);
				res = 1;
			}
		} else if (IS_STP_OR_SATA_ATAPI(pDevice)) {
			res = min(queue_depth, ARCSAS_MAX_CMD_PERDEV);
			if (sdev->tagged_supported)
				scsi_adjust_queue_depth(sdev, scsi_get_tag_type(sdev), res);
			else {
				scsi_adjust_queue_depth(sdev, 0, 1);
				res = 1;
			}			
		}
		osm_printk("queue_depth of scsi id:lun(%d:%d) has changed as %d\n", res, sdev->id, sdev->lun);
		return res;
	}
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 11)
	static int arcsas_change_queue_depth(struct scsi_device *sdev, int queue_depth)
	{
		unsigned short entry_index;
		PDomain_Device pDevice = NULL;
		int res = 0;
		if (sdev->lun) {
			unsigned char i;
			entry_index = MAKE_DEVICE_ENTRY_INDEX(sdev->id, sdev->lun);
			for (i = 0; i <= ARCSAS_MAX_TARGETLUN; i++) {
				if (acb_addr->DevicesLun[i].EntryIndex == entry_index) {/*this value is (target_lun,target_id)*/
					pDevice = &acb_addr->DevicesLun[i];
					break;
				}
			}
		} else {
			entry_index = Get_DevicesExpandersPMs_EntryIndex(acb_addr, sdev->id);
			if (entry_index == ID_NOT_MAPPED) {
				pDevice = NULL;/*hot unplug, device is disconnected*/
			} else {
				pDevice = &acb_addr->Devices[entry_index];
			}
		}
		if (pDevice == NULL) {
			osm_printk("fail to change queue_depth of scsi id:lun(%d:%d)\n", sdev->id, sdev->lun);
			return -EOPNOTSUPP;
		}
		if (IS_SSP(pDevice)) {
			res = min(queue_depth, SAS_MAX_QD);
			if (sdev->tagged_supported)
				scsi_adjust_queue_depth(sdev, scsi_get_tag_type(sdev), res);
			else {
				scsi_adjust_queue_depth(sdev, 0, 1);
				res = 1;
			}
		} else if (IS_STP_OR_SATA_ATAPI(pDevice)) {
			res = min(queue_depth, ARCSAS_MAX_CMD_PERDEV);
			if (sdev->tagged_supported)
				scsi_adjust_queue_depth(sdev, scsi_get_tag_type(sdev), res);
			else {
				scsi_adjust_queue_depth(sdev, 0, 1);
				res = 1;
			}			
		}
		osm_printk("queue_depth of scsi id:lun(%d:%d) has changed as %d\n", res, sdev->id, sdev->lun);
		return res;
	}
#else
	static ssize_t arcsas_change_queue_depth(struct device *dev, const char *buf, size_t count)
	{
		int queue_depth;
		struct scsi_device *sdev = to_scsi_device(dev);
		unsigned short entry_index;
		PDomain_Device pDevice;
		int res;

		if (sdev->lun) {
			unsigned char i;
			entry_index = MAKE_DEVICE_ENTRY_INDEX(sdev->id, sdev->lun);
			for (i = 0; i <= ARCSAS_MAX_TARGETLUN; i++) {
				if (acb_addr->DevicesLun[i].EntryIndex == entry_index) {/*this value is (target_lun,target_id)*/
					pDevice = &acb_addr->DevicesLun[i];
					break;
				}
			}
		} else {
			entry_index = Get_DevicesExpandersPMs_EntryIndex(acb_addr, sdev->id);
			if (entry_index == ID_NOT_MAPPED) {
				pDevice = NULL;/*hot unplug, device is disconnected*/
			} else {
				pDevice = &acb_addr->Devices[entry_index];
			}
		}
		if (pDevice == NULL) {
			osm_printk("fail to change queue_depth of scsi id:lun(%d:%d)\n", sdev->id, sdev->lun);
			return -EINVAL;
		}
		queue_depth = simple_strtoul(buf, NULL, 0);
		if (IS_SSP(pDevice)) {
			res = min(queue_depth, SAS_MAX_QD);
			if (sdev->tagged_supported)
				scsi_adjust_queue_depth(sdev, scsi_get_tag_type(sdev), res);
			else {
				scsi_adjust_queue_depth(sdev, 0, 1);
				res = 1;
			}
		} else if (IS_STP_OR_SATA_ATAPI(pDevice)) {
			res = min(queue_depth, ARCSAS_MAX_CMD_PERDEV);
			if (sdev->tagged_supported)
				scsi_adjust_queue_depth(sdev, scsi_get_tag_type(sdev), res);
			else {
				scsi_adjust_queue_depth(sdev, 0, 1);
				res = 1;
			}			
		}
		osm_printk("queue_depth of scsi id:lun(%d:%d) has changed as %d\n", res, sdev->id, sdev->lun);
		return res;
	}
	static struct device_attribute arcmsr_queue_depth_attr = 
	{
		.attr = {
			.name =	"queue_depth",
			.mode =	S_IRUSR | S_IWUSR,
		},
		.store = arcmsr_adjust_disk_queue_depth
	};
	static struct device_attribute *arcmsr_scsi_device_attr[] = 
	{
		&arcmsr_queue_depth_attr,
		NULL,
	};
#endif
/*
**********************************************************************
**
**********************************************************************
*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 19)
	static irqreturn_t arcsas_interrupt(int irq, void *dev_id)
#else
	static irqreturn_t arcsas_interrupt(int irq, void *dev_id, struct pt_regs *regs)
#endif
	{
		struct AdapterControlBlock *pACB;
		irqreturn_t retval;

		pACB = (struct AdapterControlBlock *)dev_id;
		retval = arcsas_interrupt_check(pACB);
		return retval;
	}
/*
*********************************************************************************
*********************************************************************************
*/
static struct scsi_host_template arcsas_scsi_host_template = {
	.module			= THIS_MODULE,
	.name			= "arcsas",
	.queuecommand		= arcsas_queuecommand,
	.eh_abort_handler		= arcsas_os_eh_abort_handler,
	.can_queue		= ARCSAS_MAX_OUTSTANDING_CMD -1,/*mvsas in kernel source*/
	.this_id			= -1, /* or ARCSAS_SCSI_INITIATOR_ID = 255*/
	.sg_tablesize		= ARCSAS_MAX_OS_SG_ENTRIES - 1,
	.max_sectors		= ARCSAS_MAX_CMDXFER_LENGTH / ARCSAS_SECTOR_SIZE,
	.cmd_per_lun		= ARCSAS_MAX_CMD_PERDEV,
	.change_queue_depth	= arcsas_change_queue_depth,
	.use_clustering		= ENABLE_CLUSTERING,
	.emulated			= 1,
};
/*
********************************************************************************
* idnetical with arcsas_HwFindAdapter in Windows driver, arcsas.1.00.00.00.
********************************************************************************
*/
static bool arcsas_initialize_hba(struct AdapterControlBlock *pacb, struct pci_dev *pcidev)
{
	unsigned long addr;
	unsigned long range;
	void __iomem *mem_map_base = NULL;
	void __iomem *mem_map_base_ext = NULL;

	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	pacb->vendor = pcidev->vendor;
	pacb->system_device_id = pcidev->device;
	pacb->subsystem_vendor = pcidev->subsystem_vendor;
	pacb->subsystem_device = pcidev->subsystem_device;
	osm_atomic_set(&pacb->outstandingccbcount, 0);
	osm_atomic_set(&pacb->all_device_count, 0);
	pacb->enclosure_mask = 0;
	/*refer to pCore->Mmio_Base = controller.Base_Address[MV_PCI_BAR] in Core_ModuleInitialize function of core_exp.c*/
	switch (pacb->system_device_id) {
		case DEVICE_ID_1300: {
			addr  = (unsigned long)pci_resource_start(pcidev, 2);
			range = pci_resource_len(pcidev, 2);
			if (!addr || !range)
				return false;
			if (pci_resource_flags(pcidev, 2) & IORESOURCE_MEM) {
				if (pci_resource_flags(pcidev, 2) & IORESOURCE_CACHEABLE) {
					#if ARCSAS_DBG_CONTEXT
						osm_printk("line %d Cachable I/O memory\n", __LINE__);
					#endif
					mem_map_base_ext = (void *)ioremap(addr, range);
				} else {
					#if ARCSAS_DBG_CONTEXT
						osm_printk("line %d Uncachable I/O memory\n", __LINE__);
					#endif
					mem_map_base_ext = (void *)ioremap_nocache(addr, range);
				}
			} else {
				#if ARCSAS_DBG_CONTEXT
					osm_printk("I/O port\n");
				#endif
				mem_map_base_ext = (void *)addr;
			}
			addr  = pci_resource_start(pcidev, 4);
			range = pci_resource_len(pcidev, 4);
			if (!addr || !range)
				return false;
			if (pci_resource_flags(pcidev, 4) & IORESOURCE_MEM) {
				if (pci_resource_flags(pcidev, 4) & IORESOURCE_CACHEABLE) {
					mem_map_base = ioremap(addr, range);
					#if ARCSAS_DBG_CONTEXT
						osm_printk("line %d Cachable I/O memory mem_map_base = 0x%p\n", __LINE__, mem_map_base);
					#endif
				} else {
					mem_map_base = ioremap_nocache(addr, range);
					#if ARCSAS_DBG_CONTEXT
						osm_printk("line %d Uncachable I/O memory mem_map_base = 0x%p\n", __LINE__, mem_map_base);
					#endif
				}
			} else {
				#if ARCSAS_DBG_CONTEXT
					osm_printk("I/O port\n");
				#endif
				mem_map_base_ext = (void *)addr;
			}
			if (!mem_map_base) {
				if (pci_resource_flags(pcidev, 2) & IORESOURCE_MEM) {
					osm_printk("line %d ARC1300 iounmap\n", __LINE__);
					iounmap(mem_map_base_ext);
				}
				return false;
			}
			pacb->acb_flag &= ~ACB_FLAG_DUAL_CORE;
			if (!arcsas_core_initialize(pacb)) {
				osm_printk("line %d ARC1300 iounmap\n", __LINE__);
				iounmap(mem_map_base);
				iounmap(mem_map_base_ext);
				return false;
			}
			pacb->pCore[0]->io_map_base = mem_map_base_ext;
			pacb->pCore[0]->mem_map_base = mem_map_base;
			break;
		}
		case DEVICE_ID_1320:{
			addr  = pci_resource_start(pcidev, 2);
			range = pci_resource_len(pcidev, 2);
			if (pci_resource_flags(pcidev, 2) & IORESOURCE_CACHEABLE) {
				mem_map_base_ext = ioremap(addr, range);
				#if ARCSAS_DBG_CONTEXT
					osm_printk("line %d Cachable I/O memory\n", __LINE__);
				#endif
			} else {
				mem_map_base_ext = ioremap_nocache(addr, range);
				#if ARCSAS_DBG_CONTEXT
					osm_printk("line %d Uncachable I/O memory\n", __LINE__);
				#endif
			}
			if (!mem_map_base_ext) {
				osm_printk("ioremap failed......\n");
				return false;
			}
			pacb->acb_flag |= ACB_FLAG_DUAL_CORE;
			if (!arcsas_core_initialize(pacb)) {
				osm_printk("core initialization is not finished..........\n");
				return false;
			}
			/*
			***************************************************************************************
			** pCore->mem_map_base: (SAS/SATA A Register) port0...port3 20000h-23FFFh
			** pCore->mem_map_base+0x4000: (SAS/SATA B Register) port4...port7 24000h-24FFFh
			***************************************************************************************
			*/
			pacb->pCore[0]->mem_map_base_ext = mem_map_base_ext;
			pacb->pCore[1]->mem_map_base_ext = mem_map_base_ext;
			pacb->pCore[0]->mem_map_base = NumToPtr(PtrToNum(mem_map_base_ext) + 0x20000);
			pacb->pCore[1]->mem_map_base = NumToPtr(PtrToNum(mem_map_base_ext) + 0x24000);
			break;
		}
		default:{
			_ASSERT();
		}
	}
	if (arcsas_reset_controller(pacb) == false) {
		arcsas_teardown_dma_pool(pacb);
		switch (pacb->system_device_id) {
			case DEVICE_ID_1300: {
				osm_printk("line %d ARC1300 iounmap\n", __LINE__);
				iounmap(pacb->pCore[0]->mem_map_base);
				break;
			}
			case DEVICE_ID_1320: {
				osm_printk("ARC1320 iounmap\n");
				iounmap((pacb->pCore[0]->mem_map_base) - 0x20000);
				break;
			}
			default:{
				_ASSERT();
			}
		}
		return false;
	}	
	return arcsas_init_chip(pacb);
}
/*
*********************************************************************************
*********************************************************************************
*/
static int arcsas_pci_probe(struct pci_dev *pcidev, const struct pci_device_id *id)
{
	struct Scsi_Host *host;
	struct AdapterControlBlock *pacb;
	uint8_t dev_fun;
	unsigned int retval = -ENXIO;
	PCore_Extension pCore = 0;
	unsigned int can_queue = 0;
	unsigned int max_core_num, core_index;
	int rc;
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	retval = pci_enable_device(pcidev);
	if (retval) {
		osm_printk("pci_enable_device returned fail........\n");
		goto out;
	}
	if ((pcidev->subsystem_vendor != PCI_VENDOR_ID_ARECA)) {
		osm_printk("Vendor ID=%x........\n", pcidev->vendor);
		retval = -ENODEV;
		goto out_disable_device;
	}
	pci_set_master(pcidev);
	host = scsi_host_alloc(&arcsas_scsi_host_template, sizeof(struct AdapterControlBlock));
	if (!host) {
		retval = -ENOMEM;
		goto out_disable_device;
	}
	acb_addr = pacb = (struct AdapterControlBlock *)host->hostdata;
	osm_pci_read_config_byte(pcidev, PCI_REVISION_ID, &pacb->Revision_Id);
	osm_memset(pacb, 0, sizeof(struct AdapterControlBlock));
	if (!pci_set_dma_mask(pcidev, DMA_BIT_MASK(64))) {
		rc = pci_set_consistent_dma_mask(pcidev, DMA_BIT_MASK(64));
		if (rc) {
			rc = pci_set_consistent_dma_mask(pcidev, DMA_BIT_MASK(32));
			if (rc) {
				osm_printk("64-bit DMA enable failed\n");
				goto out_host_put;
			}
		}
	} else {
		rc = pci_set_dma_mask(pcidev, DMA_BIT_MASK(32));
		if (rc) {
			osm_printk("32-bit DMA enable failed\n");
			goto out_host_put;
		}
		rc = pci_set_consistent_dma_mask(pcidev, DMA_BIT_MASK(32));
		if (rc) {
			osm_printk("32-bit consistent DMA enable failed\n");
			goto out_host_put;
		}
	}
	if (pcidev->device & 0x1300) {
		can_queue = (1U << 9) - 2;//which is corresponding to ARCSAS_MAX_OUTSTANDING_CMD;
	} else if (pcidev->device & 0x1320) {
		can_queue = (1U << 10) - 2;
	}
	dev_fun = pcidev->devfn;
	pacb->host = host;
	pacb->pcidev = pcidev;
	/*the parameters for "struct Scsi_Host" can't be omitted, or something troublesome will happen*/
	host->max_id =  ARCSAS_MAX_DISK_ID; /*If this parameter is 137, scsi layer will dispatch inquiry command to device with scsi ID = 137*/
	host->max_sectors = ARCSAS_MAX_CMDXFER_LENGTH / ARCSAS_SECTOR_SIZE;
	host->max_cmd_len = ARCSAS_MAX_CMD_LEN;    /*this is issue of 64bit LBA, over 2T byte*/
	host->sg_tablesize = ARCSAS_MAX_OS_SG_ENTRIES - 1;
	host->can_queue = can_queue; /* max simultaneous cmds */
	host->this_id = -1;
	host->max_lun = max_lun;
	pci_set_drvdata(pcidev, host);
	/*******************************************************************/
	retval = pci_request_regions(pcidev, "arcsas");
	if (retval) {
		goto out_host_put;
	}
	/*******************************************************************/
	retval = arcsas_initialize_hba(pacb, pcidev);
	if (!retval) {
		retval = -ENOMEM;
		goto out_release_regions;
	}
	#if ARCSAS_SUPPORT_SES
		arcsas_init_virtual_ses_device(pacb);
	#endif
	/*******************************************************************/
	arcsas_spin_lock_init(&pacb->doneq_lock);
	arcsas_spin_lock_init(&pacb->discoveryQ_lock);
	spin_lock_init(&pacb->host_lock);
	spin_lock_init(&pacb->event_element_lock);
	pacb->acb_flag |= ACB_FLAG_PCI_PROBE;
	max_core_num = (pacb->acb_flag & ACB_FLAG_DUAL_CORE) ? 2 : 1;
	for (core_index = 0; core_index < max_core_num; core_index++) {
		pCore = (PCore_Extension)pacb->pCore[core_index];
		pCore->State |= CORE_STATE_IDLE;
		arcsas_port_state_machine(pCore, NULL);
		do {
			arcsas_interrupt_check(pacb);
		} while (!(pCore->State & CORE_STATE_STARTED));
	}
	/*******************************************************************/
	#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 19)
		retval = request_irq(pcidev->irq, arcsas_interrupt, IRQF_SHARED, "arcsas", pacb);
	#else
		retval = request_irq(pcidev->irq, arcsas_interrupt, SA_SHIRQ, "arcsas", pacb);
	#endif
	if (retval) {
		for (core_index = 0; core_index < max_core_num; core_index++) {
			tasklet_kill(&pacb->pCore[core_index]->isr_tasklet);
		}
		goto out_release_regions;
	}
	host->irq = pcidev->irq;
	arcsas_enable_interrupt(pacb);
	#if ARCSAS_SUPPORT_SES
		arcsas_send_ses_diagnostic_for_all_enclosure(pacb);
	#endif
	retval = scsi_add_host(host, &pcidev->dev);
	if (retval) {
		goto out_free_irq;
	}
	scsi_scan_host(host);
	pacb->acb_flag &= ~ACB_FLAG_PCI_PROBE;
	if (pacb->acb_flag & ACB_FLAG_PERSISTENT_MAP_ENABLE) {
		pacb->acb_flag &= ~ACB_FLAG_ENTRY_INDEX_UPDATED;
		/* rom region 0x30000....0x30400, 8 bytes pattern for each device 8*128=1024 */
		arcsas_check_update_flash_device_map(pacb, true);
	}
	return 0;
	out_free_irq:
		free_irq(pcidev->irq, pacb);
		for (core_index = 0; core_index < max_core_num; core_index++) {
			tasklet_kill(&pacb->pCore[core_index]->isr_tasklet);
		}
		switch (pacb->system_device_id) {
			case DEVICE_ID_1300: {
				iounmap(pacb->pCore[0]->mem_map_base);
				break;
			}
			case DEVICE_ID_1320: {
				iounmap((pacb->pCore[0]->mem_map_base) - 0x20000);
				break;
			}
			default: {
				_ASSERT();
				break;
			}
		}
		arcsas_teardown_dma_pool(pacb);
	out_release_regions:
		osm_vfree(pacb->DevicesLun);
		pci_release_regions(pcidev);
	out_host_put:
		pci_set_drvdata(pcidev, NULL);
		scsi_host_put(host);
	out_disable_device:
		pci_disable_device(pcidev);
	out:
		return retval;
}
/*
*********************************************************************************
*********************************************************************************
*/
static int arcsas_module_init(void)
{
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	return pci_register_driver(&arcsas_pci_driver);
}
/*
*********************************************************************************
*********************************************************************************
*/
static void  arcsas_module_exit(void)
{
	#if ARCSAS_DBG_FUNCTION
		osm_printk("%s:\n", __func__);
	#endif
	pci_unregister_driver(&arcsas_pci_driver);
}
module_init(arcsas_module_init);
module_exit(arcsas_module_exit);

